# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import functools
import logging
import operator
from collections import defaultdict
from collections.abc import Iterable
from copy import deepcopy

import gaarf
import proto
import tenacity
from google.ads import googleads
from google.api_core import exceptions

from googleads_housekeeper.services import enums


@dataclasses.dataclass
class ExclusionResult:
    """Contains data on results of placement exclusion.

    As a result of exclusion some placements can be excluded and some can be
    associated with placement exclusion lists.

    Attributes:
        excluded_placements:
            A report containing placements that were excluded from Google Ads.
        associated_with_list_placements:
            A report containing placements that were added to placement
            negative lists.
    """
    excluded_placements: gaarf.report.GaarfReport | None = None
    associated_with_list_placements: gaarf.report.GaarfReport | None = None


@dataclasses.dataclass
class ExclusionOperations:
    """Operations to be sent to Google Ads API.

    Attributes:
        placement_exclusion_operations:
            Operations for excluding placements from Google Ads.
        shared_set_creation_operations:
            Operations for creating placement exclusion lists in Google Ads.
        campaign_set_association_operations:
            Operations for attaching placements to placement exclusion lists
            in Google Ads.
        excluded_placements:
            A report containing placements that were excluded from Google Ads.
    """
    placement_exclusion_operations: dict
    shared_set_creation_operations: dict
    campaign_set_association_operations: dict
    excluded_placements: gaarf.report.GaarfReport
    associated_with_list_placements: gaarf.report.GaarfReport


@dataclasses.dataclass
class PlacementOperation:
    """Represent exclusion operation for single placement.

    Attributes:
        customer_id:
            Google Ads account id.
        exclusion_operation:
            Operation for excluding placement from Google Ads.
        shared_set_resource_name:
            Resource name of placement exclusion list the placement can be
            associated with.
        is_associatable:
            Whether or not placement can be associated with placement exclusion
            list.
    """
    customer_id: int
    exclusion_operation: proto.message.Message
    shared_set_resource_name: str | None = None
    is_associatable: bool = False


class PlacementExcluder:
    """Handles excluding placements from Google Ads.


    Attributes:
        associable_with_negative_lists:
            Placements from this campaigns can be added to placement exclusion
            lists but campaigns cannot be associated with the list via API.
        associatable_to_negative_lists:
            Placements from this campaigns can be added to placement exclusion
            lists and campaigns can be associated with the list via API.
    """

    associable_with_negative_lists = ('VIDEO',)
    associatable_to_negative_lists = ('DISPLAY', 'SEARCH')

    def __init__(self, client: gaarf.api_clients.GoogleAdsApiClient,
                 exclusion_level: enums.ExclusionLevelEnum) -> None:
        self._client = client
        self.exclusion_level = exclusion_level

    @property
    def client(self):
        """Google Ads client to handle mutate requests."""
        return self._client.client

    @functools.cached_property
    def campaign_service(self):
        """Google Ads service for mutating campaigns."""
        return self.client.get_service('CampaignService')

    @functools.cached_property
    def campaign_set_operation(self):
        """Google Ads operation for creating campaign shared sets."""
        return self.client.get_type('CampaignSharedSetOperation')

    @functools.cached_property
    def shared_set_service(self):
        """Google Ads service for creating shared sets."""
        return self.client.get_service('SharedSetService')

    @functools.cached_property
    def shared_criterion_service(self):
        """Google Ads service for creating shared set criterion (placements)."""
        return self.client.get_service('SharedCriterionService')

    @functools.cached_property
    def campaign_shared_set_service(self):
        """Google Ads service for attaching campaigns to share set."""
        return self.client.get_service('CampaignSharedSetService')

    @functools.cached_property
    def shared_set_operation(self):
        """Google Ads operation for managing shared sets."""
        return self.client.get_type('SharedSetOperation')

    @functools.cached_property
    def criterion_service(self):
        """Google Ads service for adding placement as a negative criterion."""
        if self.exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
            return self.client.get_service('AdGroupCriterionService')
        if self.exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
            return self.client.get_service('CampaignCriterionService')
        if self.exclusion_level == enums.ExclusionLevelEnum.ACCOUNT:
            return self.client.get_service('CustomerNegativeCriterionService')
        return None

    @functools.cached_property
    def mutate_operation(self):
        """Google Ads operation for changing negative criteria."""
        if self.exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
            return self.criterion_service.mutate_ad_group_criteria
        if self.exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
            return self.criterion_service.mutate_campaign_criteria
        if self.exclusion_level == enums.ExclusionLevelEnum.ACCOUNT:
            return self.criterion_service.mutate_customer_negative_criteria
        return None

    @functools.cached_property
    def criterion_operation(self):
        """Google Ads operation for managing negative criteria."""
        if self.exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
            return self.client.get_type('AdGroupCriterionOperation')
        if self.exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
            return self.client.get_type('CampaignCriterionOperation')
        if self.exclusion_level == enums.ExclusionLevelEnum.ACCOUNT:
            return self.client.get_type('CustomerNegativeCriterionOperation')
        return None

    @functools.cached_property
    def entity_id(self):
        """Specified resource id (ad_group, campaign) for building path."""
        if self.exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
            return 'ad_group_id'
        if self.exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
            return 'campaign_id'
        return None

    def exclude_placements(
        self,
        to_be_excluded_placements: gaarf.report.GaarfReport,
        placement_exclusion_lists: dict[str, str],
    ) -> ExclusionResult:
        """Perform exclusion of placements from Google Ads.

        Some placement might be directly excluded while other are associated
        with a negative placement list.

        Args:
            to_be_excluded_placements:
                Report containing placements to be excluded.
            placement_exclusion_lists:
                Mapping between placement exclusion list name and
                its resource_name.
        Returns:
            Results of exclusion.
        """
        exclusion_operations = (
            self._create_exclusion_operations(to_be_excluded_placements,
                                              placement_exclusion_lists))
        for customer_id, operations in (
                exclusion_operations.shared_set_creation_operations.items()):
            try:
                if operations:
                    self._add_placements_to_shared_set(customer_id, operations)
                    logging.info(
                        'Added %d placements to shared_set for %d account',
                        len(operations), customer_id)
            except Exception as e:
                logging.error(e)
        for customer_id, operations in (
                exclusion_operations.placement_exclusion_operations.items()):
            try:
                if operations:
                    self._exclude(customer_id, operations)
                    logging.info('Excluded %d placements from account %s',
                                 len(operations), customer_id)
            except Exception as e:
                logging.error(e)
        logging.info('%d placements was excluded',
                     len(exclusion_operations.excluded_placements))
        if exclusion_operations.campaign_set_association_operations:
            operations = self._create_campaign_set_operations(
                customer_id,
                exclusion_operations.campaign_set_association_operations)
            self._add_campaigns_to_shared_set(customer_id, operations)
        if excluded_placements := exclusion_operations.excluded_placements:
            excluded_placements = functools.reduce(
                operator.add, exclusion_operations.excluded_placements)
        if (associated_with_list_placements :=
                exclusion_operations.associated_with_list_placements):
            associated_with_list_placements = functools.reduce(
                operator.add, associated_with_list_placements)
        return ExclusionResult(
            excluded_placements=excluded_placements,
            associated_with_list_placements=associated_with_list_placements)

    def _create_exclusion_operations(
        self,
        placements: gaarf.report.GaarfReport,
        placement_exclusion_lists: dict[str, str],
    ) -> ExclusionOperations:
        """Create exclusion operations to be send to Google Ads API.

        Args:
            to_be_excluded_placements:
                Report containing placements to be excluded.
            placement_exclusion_lists:
                Mapping between placement exclusion list name and
                its resource_name.
        Returns:
            Mutate operations to be send to Google Ads API.
        """
        operations_mapping: dict[str, list] = defaultdict(list)
        excluded_placements: list[gaarf.report.GaarfReport] = []
        shared_set_operations_mapping: dict[str, list] = defaultdict(list)
        associated_with_list_placements: list[gaarf.report.GaarfReport] = []
        campaign_set_mapping: dict[str, int] = {}
        for placement_info in placements:
            if placement_info.allowlisted or placement_info.excluded_already:
                continue
            placement_operation, relavant_placement_info = (
                self._create_placement_operation(placement_info,
                                                 placement_exclusion_lists))
            if shared_set := placement_operation.shared_set_resource_name:
                shared_set_operations_mapping[
                    placement_operation.customer_id].append(
                        placement_operation.exclusion_operation)
                associated_with_list_placements.append(relavant_placement_info)
                if placement_operation.is_associatable:
                    campaign_set_mapping[
                        shared_set] = placement_info.campaign_id
            else:
                operations_mapping[placement_operation.customer_id].append(
                    placement_operation.exclusion_operation)
                excluded_placements.append(relavant_placement_info)
        return ExclusionOperations(
            placement_exclusion_operations=operations_mapping,
            shared_set_creation_operations=shared_set_operations_mapping,
            campaign_set_association_operations=campaign_set_mapping,
            excluded_placements=excluded_placements,
            associated_with_list_placements=associated_with_list_placements)

    def _create_placement_operation(
        self,
        placement_info: gaarf.report.GaarfRow,
        placement_exclusion_lists: dict[str, str] | None = None
    ) -> tuple[PlacementOperation, gaarf.report.GaarfReport]:
        """Create exclusion operation for a single placement.

        Args:
            placement_info:
                Row from report containing placements to be excluded.
            placement_exclusion_lists:
                Mapping between placement exclusion list name and
                its resource_name.
        Returns:
            Tuple containing exclusion operation and subset of placement_info
            with information on placement.
        """
        entity_criterion = None
        shared_set_resource_name = None
        is_associatable = False
        if (self.exclusion_level in (enums.ExclusionLevelEnum.CAMPAIGN,
                                     enums.ExclusionLevelEnum.AD_GROUP) and
                placement_info.campaign_type
                in self.associable_with_negative_lists):
            if shared_set_resource_name := self._create_shared_set(
                    placement_info.customer_id, placement_info.campaign_id,
                    placement_exclusion_lists):
                shared_criterion_operation = self.client.get_type(
                    'SharedCriterionOperation')
                entity_criterion = shared_criterion_operation.create
                entity_criterion.shared_set = shared_set_resource_name
            if placement_info.campaign_type in self.associatable_to_negative_lists:
                is_associatable = True

        if (placement_info.placement_type ==
                enums.PlacementTypeEnum.MOBILE_APPLICATION.name):
            app_id = self._format_app_id(placement_info.placement)
        if not entity_criterion:
            entity_criterion = self.criterion_operation.create
        # Assign specific criterion
        if placement_info.placement_type == (
                enums.PlacementTypeEnum.WEBSITE.name):
            entity_criterion.placement.url = placement_info.placement
        if (placement_info.placement_type ==
                enums.PlacementTypeEnum.MOBILE_APPLICATION.name):
            entity_criterion.mobile_application.app_id = app_id
        if placement_info.placement_type == (
                enums.PlacementTypeEnum.YOUTUBE_VIDEO.name):
            entity_criterion.youtube_video.video_id = placement_info.placement
        if (placement_info.placement_type ==
                enums.PlacementTypeEnum.YOUTUBE_CHANNEL.name):
            entity_criterion.youtube_channel.channel_id = (
                placement_info.placement)
        if not shared_set_resource_name:
            if self.exclusion_level != enums.ExclusionLevelEnum.ACCOUNT:
                entity_criterion.negative = True
            if self.exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
                entity_criterion.ad_group = self.criterion_service.ad_group_path(
                    placement_info.customer_id,
                    placement_info.get(self.entity_id))
            elif self.exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
                entity_criterion.campaign = self.criterion_service.campaign_path(
                    placement_info.customer_id,
                    placement_info.get(self.entity_id))
        if shared_set_resource_name:
            operation = deepcopy(shared_criterion_operation)
        else:
            operation = deepcopy(self.criterion_operation)
        placement_operation = PlacementOperation(
            customer_id=placement_info.customer_id,
            exclusion_operation=operation,
            shared_set_resource_name=shared_set_resource_name,
            is_associatable=is_associatable)
        revelant_placement_info = gaarf.report.GaarfReport(
            results=[placement_info.data],
            column_names=placement_info.column_names)
        return placement_operation, revelant_placement_info

    def _create_shared_set(
        self,
        customer_id: int,
        campaign_id: int,
        placement_exclusion_lists: dict[str, str],
        base_share_set_name: str = 'CPR Negative placements list - Campaign:'
    ) -> str | None:
        """Create placement exclusion list in Google Ads API.

        Args:
            customer_id:
                Google Ads account number.
            campaign_id:
                Id of campaign that will be associated with the placement
                exclusion list.
            placement_exclusion_lists:
                Mapping between already existing placements exclusion lists and
                its resource names.
            base_share_set_name:
                Common name for all placement exclusion lists.

        Returns:
            Resource name of extracted or created placement exclusion list
            resource name.
        """
        name = f'{base_share_set_name} {campaign_id}'
        if name in placement_exclusion_lists:
            return placement_exclusion_lists[name]
        shared_set = self.shared_set_operation.create
        shared_set.name = name
        shared_set.type_ = (
            self.client.enums.SharedSetTypeEnum.NEGATIVE_PLACEMENTS)

        operation = deepcopy(self.shared_set_operation)
        try:
            shared_set_response = self.shared_set_service.mutate_shared_sets(
                customer_id=str(customer_id), operations=[operation])
            shared_set_resource_name = shared_set_response.results[
                0].resource_name
            logging.debug('Created shared set "%s".', shared_set_resource_name)
            return shared_set_resource_name
        except googleads.errors.GoogleAdsException:
            logging.debug('Shared set "%s" already exists.', name)
            return None

    def _add_placements_to_shared_set(self, customer_id: int,
                                      operations: list) -> None:
        """Adds placements to placement exclusion list.

        Args:
            customer_id:
                Google Ads account number.
            operations:
                Shared set attachment operations for a given customer_id.
        """
        if not isinstance(operations, Iterable):
            operations = [operations]
        try:
            for attempt in tenacity.Retrying(
                    retry=tenacity.retry_if_exception_type(
                        exceptions.InternalServerError),
                    stop=tenacity.stop_after_attempt(3),
                    wait=tenacity.wait_exponential()):
                with attempt:
                    self.shared_criterion_service.mutate_shared_criteria(
                        customer_id=str(customer_id), operations=operations)
        except tenacity.RetryError as retry_failure:
            logging.error(
                "Cannot add placements to exclusion list for account '%s' %d times",
                customer_id, retry_failure.last_attempt.attempt_number)

    def _create_campaign_set_operations(self, customer_id,
                                        campaign_set_mapping: dict) -> list:
        """Create operations for adding placements to placement exclusion lists.

        Args:
            customer_id:
                Google Ads account number.
            campaign_set_mapping:
               Mapping between shared_set resource_name and campaign_id.

        Returns:
            Operations for adding campaigns to placement exclusion lists.
        """
        campaign_set = self.campaign_set_operation.create
        operations = []
        for shared_set, campaign_id in campaign_set_mapping.items():
            campaign_set.campaign = self.campaign_service.campaign_path(
                customer_id, campaign_id)
            campaign_set.shared_set = shared_set
            operation = deepcopy(self.campaign_set_operation)
            operations.append(operation)
        return operations

    def _add_campaigns_to_shared_set(self, customer_id: str,
                                     operations: list) -> None:
        """Adds campaigns to placement exclusion list.

        Args:
            customer_id:
                Google Ads account number.
            operations:
                Operations for adding campaigns to placement exclusion lists.
        """
        self.campaign_shared_set_service.mutate_campaign_shared_sets(
            customer_id=str(customer_id), operations=operations)

    def _format_app_id(self, app_id: str) -> str:
        """Returns app_id as acceptable negative criterion."""
        if app_id.startswith('mobileapp::'):
            criteria = app_id.split('-')
            app_id = criteria[-1]
            app_store = criteria[0].split('::')[-1]
            app_store = app_store.replace('mobileapp::1000', '')
            app_store = app_store.replace('1000', '')
            return f'{app_store}-{app_id}'
        return app_id

    def _exclude(self, customer_id: str, operations) -> None:
        """Exclude placements from Google Ads.

        Args:
            customer_id:
                Google Ads account number.
            operations:
                Operations for excluding placements from Google Ads.
        """
        if not isinstance(operations, Iterable):
            operations = [operations]
        try:
            for attempt in tenacity.Retrying(
                    retry=tenacity.retry_if_exception_type(
                        exceptions.InternalServerError),
                    stop=tenacity.stop_after_attempt(3),
                    wait=tenacity.wait_exponential()):
                with attempt:
                    self.mutate_operation(
                        customer_id=str(customer_id), operations=operations)
        except tenacity.RetryError as retry_failure:
            logging.error("Cannot exclude placements for account '%s' %d times",
                          customer_id,
                          retry_failure.last_attempt.attempt_number)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pre-build views for simplifying reads from DB."""
from __future__ import annotations

from dataclasses import asdict
from typing import Any

from googleads_housekeeper.services import unit_of_work

registry = {}


def task(task_id: int,
         uow: unit_of_work.AbstractUnitOfWork) -> dict[str, Any] | None:
    with uow:
        if uow_task := uow.tasks.get(task_id):
            return asdict(uow_task)
        return None


def tasks(uow: unit_of_work.AbstractUnitOfWork) -> list[dict[str, Any]]:
    with uow:
        uow_tasks: list[dict[str, Any]] = []
        for uow_task in uow.tasks.list():
            task_dict = uow_task.to_dict()
            task_dict['next_run'] = uow_task.next_run
            accounts = customer_ids(
                uow=uow, accounts=uow_task.customer_ids.split(','))
            task_dict['accounts'] = list(
                {a.get('account_name') for a in accounts})
            if task_executions := executions(uow_task.id, uow):
                for current_execution in sorted(
                        task_executions,
                        key=lambda x: x.get('end_time'),
                        reverse=True):
                    task_dict['placements_excluded'] = current_execution.get(
                        'placements_excluded')
                    task_dict['last_run'] = current_execution.get(
                        'end_time').strftime('%Y-%m-%d %H-%M')
                    break
            else:
                task_dict['placements_excluded'] = None
            uow_tasks.append(task_dict)
        return uow_tasks


def config(uow: unit_of_work.AbstractUnitOfWork) -> list[dict[str, Any]]:
    with uow:
        return [asdict(c) for c in uow.settings.list()]


def allowlisted_placements(uow: unit_of_work.AbstractUnitOfWork,
                           account_id: str | None = None) -> list[dict] | None:
    with uow:
        if account_id:
            if uow_allowlisted_placements := uow.allowlisting.get_by_conditions(
                {'account_id': account_id}):
                return [asdict(t) for t in uow_allowlisted_placements]
        if uow_allowlisted_placements := uow.allowlisting.list():
            return [asdict(t) for t in uow_allowlisted_placements]
    return None


def allowlisted_placements_as_tuples(
        uow: unit_of_work.AbstractUnitOfWork,
        account_id: str | None = None) -> list[tuple[str, str, str]]:
    if placements := allowlisted_placements(uow, account_id):
        return [(p.get('account_id'), p.get('type'),
                 p.get('name')) for p in placements]
    return []


def execution(execution_id: str,
              uow: unit_of_work.AbstractUnitOfWork) -> dict[str, Any] | None:
    with uow:
        if task_execution := uow.executions.get(execution_id):
            return asdict(task_execution)
    return None


def executions(
        task_id: str,
        uow: unit_of_work.AbstractUnitOfWork) -> list[dict[str, Any]] | None:
    with uow:
        if task_executions := uow.executions.get_by_condition('task', task_id):
            return [asdict(t) for t in task_executions]
    return None


def execution_details(
        task_id: str,
        execution_id: str,
        uow: unit_of_work.AbstractUnitOfWork,
        first_n_placements: int | None = None) -> list[dict[str, Any]] | None:
    with uow:
        if uow.executions.get_by_condition('task', task_id):
            if uow_execution_details := uow.execution_details.get_by_condition(
                    'execution_id', execution_id):
                results = [asdict(t) for t in uow_execution_details]
                if first_n_placements and first_n_placements < len(results):
                    results = results[:first_n_placements]
                return results
    return None


def customer_ids(uow: unit_of_work.AbstractUnitOfWork,
                 mcc_id: str | None = None,
                 accounts: list[str] | None = None) -> list[dict[str, Any]]:
    with uow:
        if mcc_id:
            uow_customer_ids = [
                asdict(r) for r in uow.customer_ids.list() if r.mcc_id == mcc_id
            ]
        elif accounts:
            uow_customer_ids = [
                asdict(r) for r in uow.customer_ids.list() if r.id in accounts
            ]
        else:
            raise ValueError('Neither mcc_id nor accounts were provided')
        return uow_customer_ids


def mcc_ids(uow: unit_of_work.AbstractUnitOfWork) -> list[dict[str, str]]:
    if registry_mcc_ids := registry.get('mcc_ids'):
        return registry_mcc_ids
    with uow:
        uow_mcc_ids = [asdict(r) for r in uow.mcc_ids.list()]
        registry['mcc_ids'] = uow_mcc_ids
        return uow_mcc_ids



from __future__ import annotations

from datetime import datetime

import pytest

from googleads_housekeeper import bootstrap
from googleads_housekeeper.adapters import notifications
from googleads_housekeeper.adapters import publisher
from googleads_housekeeper.domain import commands
from googleads_housekeeper.domain import external_parsers
from googleads_housekeeper.services import unit_of_work


class FakeGoogleAdsApiClient:
    ...


class FakeNotifications(notifications.BaseNotifications):

    def __init__(self):
        self.sent = defaultdict(list)  # type: Dict[str, List[str]]

    def publish(self, topic, event):
        self.sent[destination].append(message)


class FakePublisher(publisher.BasePublisher):

    def __init__(self):
        self.events = []  # type: Dict[str, List[str]]

    def publish(self, topic, event):
        self.events.append(event)


@pytest.fixture(scope='session')
def in_memory_sqlite_db():
    return 'sqlite:///:memory:'


@pytest.fixture(scope='session')
def fake_publisher():
    return FakePublisher()


@pytest.fixture(scope='session')
def bus(fake_publisher, in_memory_sqlite_db):
    bus = bootstrap.bootstrap(
        start_orm=True,
        ads_api_client=FakeGoogleAdsApiClient(),
        uow=unit_of_work.SqlAlchemyUnitOfWork(in_memory_sqlite_db),
        publish_service=fake_publisher)
    new_website_info = external_parsers.website_parser.WebsiteInfo(
        placement='example.com',
        title='fun games',
        description='millions of fun games',
        keywords='browser games, free games',
        is_processed=True,
        last_processed_time=datetime.now())
    new_channel_info = external_parsers.youtube_data_parser.ChannelInfo(
        placement='1kdjf0skdjfw0dsdf',
        title='Gardening and Games',
        description='Everything about Games and Gardens',
        country='US',
        viewCount=1000,
        subscriberCount=100,
        videoCount=10,
        topicCategories='Gardening,Games',
        last_processed_time=datetime.now(),
        is_processed=True)
    new_video_info = external_parsers.youtube_data_parser.VideoInfo(
        placement='jojoh',
        title='Gardening and Games Vol. 2',
        description='The second volumes of the Gardening and Games series',
        defaultLanguage='en',
        defaultAudioLanguage='en',
        commentCount=10,
        favouriteCount=10,
        likeCount=10,
        viewCount=1000,
        madeForKids=True,
        topicCategories='Gardening,Games',
        tags='#multiplayer,#mro,#garden',
        last_processed_time=datetime.now(),
        is_processed=True)

    bus.handle(commands.SaveChannelInfo(new_channel_info))
    bus.handle(commands.SaveVideoInfo(new_video_info))
    bus.handle(commands.SaveWebsiteInfo(new_website_info))
    return bus



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for extracting meta data about YouTube channels and videos."""
from __future__ import annotations

import collections
import dataclasses
import os
from typing import Any
from typing import Literal

from googleapiclient.discovery import build

from googleads_housekeeper.domain.external_parsers import base_parser


class YouTubeDataConnector:
    """Specifies connections to YouTube.

    Attributes:
        service: Build YouTube service.
    """

    def __init__(self,
                 api_version: str = 'v3',
                 developer_key: str = os.getenv('YOUTUBE_DATA_API_KEY')):
        """Initializes YouTubeDataConnector.

        Args:
            api_version: Version of YouTube Data API.
            developer_key: API key to access YouTube Data API.
        Returns:
            Description of return.
        """
        self.service = build('youtube', api_version, developerKey=developer_key)

    def get_response(self, type_: Literal['videos', 'channels'], elements: str,
                     placements: collections.abc.Sequence[str]):
        """Gets API response for a correct type.

        Args:
            type_: Service type.
            elements: Elements to return in response.
            placements: Identifiers for videos/channels.
        Returns:
            Description of return.
        """
        if type_ == 'videos':
            service = self.service.videos()
        elif type_ == 'channels':
            service = self.service.channels()
        else:
            raise ValueError(f'Unsupported resource {type_}')
        return service.list(part=elements, id=placements).execute()


@dataclasses.dataclass
class ChannelInfo(base_parser.EntityInfo):
    """Contains meta information of YouTube Channel.

    Attributes:
        placement: Id of YouTube channel.
        title: Title of the channel.
        description: Channel description.
        viewCount: Number of channel views.
        subscriberCount: Number of channel subscribers.
        videoCount: Number of videos on the channel.
        topicCategories: Categories associated with the channel.
        is_processed: Whether YouTube channel was successfully parsed.
    """
    placement: str
    title: str | None = None
    description: str | None = None
    country: str | None = None
    viewCount: int = 0
    subscriberCount: int = 0
    videoCount: int = 0
    topicCategories: str = ''
    is_processed: bool = True


@dataclasses.dataclass
class VideoInfo(base_parser.EntityInfo):
    """Contains meta information of YouTube Video.

    Attributes:
        placement: Id of YouTube video.
        title: Title of the video.
        description: Channel description.
        defaultLanguage: Language of the video.
        defaultAudioLanguage: Audio language of the video.
        commentCount: Number of video comments.
        favouriteCount: Number of times video is added to favorites.
        likeCount: Number of video likes.
        viewCount: Number of videos views.
        madeForKids: Whether is the video is made for kids.
        tags: Tags associated with the video.
        topicCategories: Categories associated with the video.
        is_processed: Whether YouTube video was successfully parsed.
    """
    placement: str
    title: str | None = None
    description: str | None = None
    defaultLanguage: str | None = None
    defaultAudioLanguage: str | None = None
    commentCount: int = 0
    favouriteCount: int = 0
    likeCount: int = 0
    viewCount: int = 0
    madeForKids: bool = False
    tags: str = ''
    topicCategories: str = ''
    is_processed: bool = True


class ChannelInfoParser(base_parser.BaseParser):
    """Performs YouTube channel parsing."""

    def __init__(
            self,
            data_connector: type[YouTubeDataConnector] = YouTubeDataConnector):
        """Initializes ChannelInfoParser.

        Args:
            data_connector: Connector to get data from YouTube.
        """
        self.data_connector = data_connector

    def parse(self,
              placements: collections.abc.Sequence[str]) -> list[ChannelInfo]:
        """Parses provided YouTube channels.

        Args:
            placements: Sequence of YouTube channel ids.
        Returns:
            Parsed information in ChannelInfo format.
        """
        response = self.data_connector().get_response(
            'channels', 'id,snippet,statistics,topicDetails', placements)
        if not (items := response.get('items')):
            return [
                ChannelInfo(placement=channel_id, is_processed=False)
                for channel_id in placements
            ]
        results: list[ChannelInfo] = []
        for item in items:
            if snippet := item.get('snippet'):
                title = snippet.get('title')
                description = snippet.get('description')
                country = snippet.get('country')
            else:
                title = None
                description = None
                country = None
            if statistics := item.get('statistics'):
                subscriberCount = safe_cast(int,
                                            statistics.get('subscriberCount'))
                viewCount = safe_cast(int, statistics.get('viewCount'))
                videoCount = safe_cast(int, statistics.get('videoCount'))
            else:
                subscriberCount = 0
                viewCount = 0
                videoCount = 0
            topics = parse_topic_details(item.get('topicDetails'))
            results.append(
                ChannelInfo(
                    placement=item.get('id'),
                    title=title,
                    description=description,
                    country=country,
                    viewCount=viewCount,
                    subscriberCount=subscriberCount,
                    videoCount=videoCount,
                    topicCategories=topics))
        return results


class VideoInfoParser(base_parser.BaseParser):
    """Performs YouTube video parsing."""

    def __init__(
            self,
            data_connector: type[YouTubeDataConnector] = YouTubeDataConnector):
        """Initializes VideoInfoParser.

        Args:
            data_connector: Connector to get data from YouTube.
        """
        self.data_connector = data_connector

    def parse(self,
              placements: collections.abc.Sequence[str]) -> list[VideoInfo]:
        """Parses provided YouTube videos.

        Args:
            placements: Sequence of YouTube video ids.
        Returns:
            Parsed information in VideoInfo format.
        """
        response = self.data_connector().get_response(
            'videos',
            'id,status,snippet,statistics,contentDetails,topicDetails',
            placements)
        if not (items := response.get('items')):
            return [
                VideoInfo(placement=video_id, is_processed=False)
                for video_id in placements
            ]
        results: list[VideoInfo] = []
        for item in items:
            if snippet := item.get('snippet'):
                title = snippet.get('title')
                description = snippet.get('description')
                defaultLanguage = snippet.get('defaultLanguage')
                defaultAudioLanguage = snippet.get('defaultAudioLanguage')
                tags = snippet.get('tags')
            else:
                title = None
                description = None
                defaultLanguage = None
                defaultAudioLanguage = None
                tags = ''
            if statistics := item.get('statistics'):
                commentCount = safe_cast(int, statistics.get('commentCount'))
                favouriteCount = safe_cast(int,
                                           statistics.get('favouriteCount'))
                likeCount = safe_cast(int, statistics.get('likeCount'))
                viewCount = safe_cast(int, statistics.get('viewCount'))
            else:
                commentCount = 0
                favouriteCount = 0
                likeCount = 0
                viewCount = 0
            if status := item.get('status'):
                madeForKids = safe_cast(bool, status.get('madeForKids'))
            else:
                madeForKids = False
            topics = parse_topic_details(item.get('topicDetails'))
            if tags:
                tags = ','.join(tags)
            results.append(
                VideoInfo(
                    placement=item.get('id'),
                    title=title,
                    description=description,
                    defaultLanguage=defaultLanguage,
                    defaultAudioLanguage=defaultAudioLanguage,
                    commentCount=commentCount,
                    favouriteCount=favouriteCount,
                    likeCount=likeCount,
                    viewCount=viewCount,
                    madeForKids=madeForKids,
                    tags=tags,
                    topicCategories=topics))
        return results


def parse_topic_details(
        topic_details: collections.abc.Mapping[str, Any] | None) -> str:
    """Converts Wikipedia URLs to topic names."""
    if not topic_details:
        return ''
    if not (topic_categories := topic_details.get('topicCategories')):
        return ''
    return ','.join(list({topic.split('/')[-1] for topic in topic_categories}))


def safe_cast(callable_, value: str | None):
    if value:
        return callable_(value)
    return callable_()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for performing parsing of external entities.

External entity represents a YouTube Channel or Video, Webpage,
Mobile Application or anything that requires connection to external
(non Google Ads) system in order to get information from it.
"""
from __future__ import annotations

import dataclasses
import itertools
import logging

from gaarf import report

from googleads_housekeeper.adapters import repository
from googleads_housekeeper.domain import external_parsers
from googleads_housekeeper.domain.core import exclusion_specification
from googleads_housekeeper.services import unit_of_work

PARSER_MAPPING = {
    'youtube_channel_info':
        external_parsers.youtube_data_parser.ChannelInfoParser,
    'youtube_video_info':
        external_parsers.youtube_data_parser.VideoInfoParser,
    'website_info':
        external_parsers.website_parser.WebSiteParser,
}


@dataclasses.dataclass
class ParseOptions:
    """Specifies options regarding parsing behaviour.

    Attributes:
        save_to_db: Whether to save parsing results to DB.
        batch_size: Maximum number of entities in response.
    """
    save_to_db: bool = True
    batch_size: int = 50


class ExternalEntitiesParser:
    """Performance parsing of external entities (YouTube, Websites, etc.).

    Attributes:
        uow: Unit of work to handle transaction.
    """

    def __init__(self, uow: unit_of_work.AbstractUnitOfWork) -> None:
        """Initializes an instance of ExternalEntitiesParser.

        Args:
            uow: Unit of work to handle transaction.
        """
        self.uow = uow

    def parse_specification_chain(
        self,
        entities: report.GaarfReport,
        specification: exclusion_specification.ExclusionSpecification,
        parse_options: ParseOptions = ParseOptions()
    ) -> None:
        """Performance parsing of entities via all parsers in specification.

        Args:
            entities:
                Report containing entities that should be parsed.
            specifications:
                All possible non Ads specifications. Entities within a report
                can be parsed via different parsers (YouTube, Website, etc.)
            parse_options:
                Options for performing parsing.
        """
        with self.uow as uow:
            for specification_entries in specification.specifications:
                for specification_entry in specification_entries:
                    self._parse_via_external_parser(
                        entities=entities,
                        specification_entry=specification_entry,
                        uow=uow,
                        parse_options=parse_options)

    def _parse_via_external_parser(self, entities: report.GaarfReport,
                                   specification_entry: exclusion_specification
                                   .BaseExclusionSpecificationEntry,
                                   uow: unit_of_work.AbstractUnitOfWork,
                                   parse_options: ParseOptions) -> None:
        """Performance parsing of entities via a single parser in specification.

        Args:
            entities:
                Report containing entities that should be parsed.
            specification:
                Single non Ads specifications. Entities will be parsed via
                a single parsers (YouTube, Website, etc.
            uow:
                Unit of work to handle transaction.
            parse_options:
                Options for performing parsing.
        """
        repo = getattr(uow, specification_entry.repository_name)
        parser = PARSER_MAPPING.get(specification_entry.repository_name)
        if not_parsed_entities := self._get_not_parsed_entities(
                entities, specification_entry, repo):
            i = 0
            while batch := list(
                    itertools.islice(not_parsed_entities, i,
                                     i + parse_options.batch_size)):
                parsed_entities_info = parser().parse(batch)
                i += parse_options.batch_size
                if parse_options.save_to_db:
                    for parsed_placement in parsed_entities_info:
                        logging.debug('saving placement: %s to db',
                                      parsed_placement)
                        repo.add(parsed_placement)
                    uow.commit()

    def _get_not_parsed_entities(
            self, entities: report.GaarfReport,
            specification_entry: exclusion_specification
        .BaseExclusionSpecificationEntry,
            repo: repository.AbstractRepository) -> list[str]:
        """Get all entities not parsed and saved in DB.

        Args:
            entities:
                Report containing entities that should be parsed.
            specification_entry:
                Single non Ads specifications. Entities will be parsed via
                a single parsers (YouTube, Website, etc.).
            repo:
                Repository to get data from.
        Returns:
            Non-parsed placements.
        """
        not_parsed_entities: list[str] = []
        for entity_info in entities:
            if (entity_info.placement_type ==
                    specification_entry.corresponding_placement_type.name):

                if not repo.get_by_conditions({
                        'placement': entity_info.placement,
                }):
                    not_parsed_entities.append(str(entity_info.placement))
        return not_parsed_entities



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles interacting with Google Ads API."""
from __future__ import annotations

from collections.abc import Sequence

import gaarf
import numpy as np

from googleads_housekeeper.domain.core import exclusion_specification
from googleads_housekeeper.domain.core import task
from googleads_housekeeper.domain.placement_handler import entities
from googleads_housekeeper.services import enums


class PlacementFetcher:
    """Handles fetching of placement data from Google Ads API.

    Attributes:
        report_fetcher: An instance of AdsReportFetcher to handle API requests.
    """

    def __init__(self,
                 report_fetcher: gaarf.query_executor.AdsReportFetcher) -> None:
        """Initializes PlacementFetcher.

        Args:
            report_fetcher:
                An instance of AdsReportFetcher to handle API requests.
        """
        self.report_fetcher = report_fetcher

    def get_placements_for_account(
        self,
        account: str,
        task_obj: task.Task,
    ) -> gaarf.report.GaarfReport | None:
        """Gets placement data from a given account based on task definition.

        Args:
            account: Google Ads account to get data from.
            task_obj: A task defining parameters of fetching.
        Returns:
            Report with fetched placement data.
        """
        specification = (
            exclusion_specification.ExclusionSpecification.from_expression(
                task_obj.exclusion_rule))
        runtime_options = specification.define_runtime_options()
        if not (placements := self._get_placement_performance_data(
                account, task_obj)):
            return None
        if runtime_options.is_conversion_query and (
                placements_by_conversion_name :=
                self._get_placement_conversion_split_data(account, task_obj)):
            conversion_split_exclusion_specification = (
                exclusion_specification.ExclusionSpecification(
                    specifications=[runtime_options.conversion_rules]))
            placements_by_conversion_name = (
                conversion_split_exclusion_specification.apply_specifications(
                    placements_by_conversion_name))
            placements = self._join_conversion_split(
                placements, placements_by_conversion_name,
                runtime_options.conversion_name)
        return self._aggregate_placements(placements, task_obj.exclusion_level)

    def _get_placement_performance_data(
            self, account: str,
            task_obj: task.Task) -> gaarf.report.GaarfReport:
        """Gets placement performance data (clicks, impressions, etc.).

        Args:
            account: Google Ads account to get data from.
            task_obj: A task defining parameters of fetching.
        Returns:
            Report with fetched placement data.
        """
        placements = self._get_placement_data(
            task_obj=task_obj, account=account)
        if 'YOUTUBE_VIDEO' in task_obj.placement_types:
            if (detail_placements := self._get_placement_data(
                    task_obj=task_obj,
                    account=account,
                    placement_level_granularity='detail_placement_view')):
                placements = placements + detail_placements
        return placements

    def _get_placement_conversion_split_data(
            self, account: str,
            task_obj: task.Task) -> gaarf.report.GaarfReport:
        """Gets placement conversion split data (conversion_name, conversions).

        Args:
            account: Google Ads account to get data from.
            task_obj: A task defining parameters of fetching.
        Returns:
            Report with fetched placement conversion split data.
        """
        placements = self._get_placement_data(
            task_obj=task_obj,
            account=account,
            query_class=entities.PlacementsConversionSplit)
        if 'YOUTUBE_VIDEO' in task_obj.placement_types:
            if (detail_placements := self._get_placement_data(
                    task_obj=task_obj,
                    account=account,
                    query_class=entities.PlacementsConversionSplit)):
                placements = placements + detail_placements
        return placements

    def get_already_excluded_placements(
            self, account: str,
            exclusion_level: enums.ExclusionLevelEnum) -> dict[str, list[str]]:
        """Fetches negative placement criteria from Google Ads account.

        Negative criteria are fetched based on specified exclusion level (
        ACCOUNT, CAMPAIGN, AD_GROUP).

        Args:
            account: Google Ads account to get data from.
            exclusion_level: Level of exclusion (ACCOUNT, CAMPAIGN, AD_GROUP).
        Returns:
            Mapping between entity_id from specified exclusion level to all
            negative criteria ids.
        """
        already_excluded_placements = self.report_fetcher.fetch(
            entities.AlreadyExcludedPlacements(exclusion_level), account)
        if not already_excluded_placements:
            return {}
        for row in already_excluded_placements:
            row['placement'] = (
                row.website_url or row.app_id or row.video_id or
                row.channel_id or '')
        return already_excluded_placements.to_dict(
            key_column='level_id',
            value_column='placement',
            value_column_output='list')

    def get_placement_exclusion_lists(
            self, customer_ids: Sequence[int]) -> dict[str, str]:

        placement_exclusion_lists = self.report_fetcher.fetch(
            entities.PlacementExclusionLists(), customer_ids)
        return placement_exclusion_lists.to_dict(
            key_column='name',
            value_column='resource_name',
            value_column_output='scalar')

    def _get_placement_data(
        self,
        task_obj: task.Task,
        account: str,
        query_class: gaarf.base_query.BaseQuery = entities.Placements,
        placement_level_granularity: str = 'group_placement_view'
    ) -> gaarf.report.Report:
        """Helper method for building query and fetching data from Ads API.

        Args:
            task_obj: A Task that contains necessary data for building query.
            account: Google Ads account to fetch data from.
            query_class: Class defining which query will be built.
            placement_level_granularity: Resource name to get data from.
        Returns:
            Report containing placement performance data.
        """
        placement_query = query_class(
            placement_types=task_obj.placement_types,
            placement_level_granularity=placement_level_granularity,
            start_date=task_obj.start_date,
            end_date=task_obj.end_date)
        return self.report_fetcher.fetch(placement_query, customer_ids=account)

    def _join_conversion_split(
            self, placements: gaarf.report.GaarfReport,
            placements_by_conversion_name: gaarf.report.GaarfReport,
            conversion_name: str) -> gaarf.report.GaarfReport:
        """Joins placements performance data with its conversion split data.

        Args:
            placements:
                Report with placement performance data.
            placements_by_conversion_name:
                Report with placements conversion split data.
            conversion_name:
                Conversion_name(s) that should be used to create a dedicated
                column in joined report.

        Returns:
            New report with extra conversion specific columns.
        """
        conversion_names = conversion_name.replace('"', '').split(',')
        placements_by_conversion_name = placements_by_conversion_name.to_pandas(
        )
        final_report_values = []
        for row in placements:
            is_conversion_row = (
                placements_by_conversion_name.ad_group_id == row.ad_group_id
            ) & (placements_by_conversion_name.placement == row.placement) & (
                placements_by_conversion_name.conversion_name.isin(
                    conversion_names))
            conversion_row = placements_by_conversion_name.loc[
                is_conversion_row]
            data = list(row.data)
            if not (conversions := sum(conversion_row['conversions'].values)):
                conversions = 0.0
            if not (all_conversions := sum(
                    conversion_row['all_conversions'].values)):
                all_conversions = 0.0
            data.extend([conversion_name, conversions, all_conversions])
            final_report_values.append(data)
        columns = list(placements.column_names)
        columns.extend(['conversion_name', 'conversions_', 'all_conversions_'])
        return gaarf.report.GaarfReport(
            results=final_report_values, column_names=columns)

    def _aggregate_placements(
            self,
            placements: gaarf.report.GaarfReport,
            exclusion_level: str | enums.ExclusionLevelEnum,
            perform_relative_aggregations: bool = True
    ) -> gaarf.report.GaarfReport:
        """Aggregates placements to a desired exclusion_level.

        By default Placements report returned on Ad Group level, however exclusion
        can be performed on Campaign, Account and MCC level. By aggregating report
        to a desired level exclusion specification can be property applied to
        identify placements that should be excluded.

        Args:
            placements:
                Report with placement related metrics.
            exclusion_level:
                Desired level of aggregation.
            perform_relative_aggregations:
                Whether or not calculate relative metrics (CTR, CPC, etc.)
        Returns:
            Updated report aggregated to desired exclusion level.
        """
        if not isinstance(exclusion_level, enums.ExclusionLevelEnum):
            exclusion_level = getattr(enums.ExclusionLevelEnum, exclusion_level)
        base_groupby = [
            'placement', 'placement_type', 'name', 'criterion_id', 'url'
        ]
        aggregation_dict = dict.fromkeys([
            'clicks',
            'impressions',
            'cost',
            'conversions',
            'video_views',
            'interactions',
            'all_conversions',
            'view_through_conversions',
        ], 'sum')
        relative_aggregations_dict = {
            'ctr': ['clicks', 'impressions'],
            'avg_cpc': ['cost', 'clicks'],
            'avg_cpm': ['cost', 'impressions'],
            'avg_cpv': ['cost', 'video_views'],
            'video_view_rate': ['video_views', 'impressions'],
            'interaction_rate': ['interactions', 'clicks'],
            'conversions_from_interactions_rate': [
                'conversions', 'interactions'
            ],
            'cost_per_conversion': ['cost', 'conversions'],
            'cost_per_all_conversion': ['cost', 'all_conversions'],
            'all_conversion_rate': ['all_conversions', 'interactions'],
            'all_conversions_from_interactions_rate': [
                'all_conversions', 'interactions'
            ],
        }
        if 'conversion_name' in placements.column_names:
            base_groupby = base_groupby + ['conversion_name']
            aggregation_dict.update(
                dict.fromkeys(['conversions_', 'all_conversions_'], 'sum'))
            relative_aggregations_dict.update({
                'cost_per_conversion_': ['cost', 'conversions_'],
                'cost_per_all_conversion_': ['cost', 'all_conversions_']
            })

        aggregation_groupby = self._define_aggregation_group_by(exclusion_level)
        groupby = [
            base for base in base_groupby + aggregation_groupby
            if base in placements.column_names
        ]
        aggregations = {
            key: value
            for key, value in aggregation_dict.items()
            if key in placements.column_names
        }
        aggregated_placements = placements.to_pandas().groupby(
            groupby, as_index=False).agg(aggregations)
        if perform_relative_aggregations:
            for key, [numerator,
                      denominator] in relative_aggregations_dict.items():
                if set([numerator, denominator
                       ]).issubset(set(aggregated_placements.columns)):
                    aggregated_placements[key] = aggregated_placements[
                        numerator] / aggregated_placements[denominator]
                    if key == 'avg_cpm':
                        aggregated_placements[
                            key] = aggregated_placements[key] * 1000
                    if key == 'ctr':
                        aggregated_placements[key] = round(
                            aggregated_placements[key], 4)
                    else:
                        aggregated_placements[key] = round(
                            aggregated_placements[key], 2)
        aggregated_placements.replace([np.inf, -np.inf], 0, inplace=True)
        return gaarf.report.GaarfReport.from_pandas(aggregated_placements)

    def _define_aggregation_group_by(
            self, exclusion_level: enums.ExclusionLevelEnum) -> list[str]:
        aggregation_groupby = ['account_name', 'customer_id']
        if exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
            aggregation_groupby += [
                'campaign_id', 'campaign_name', 'campaign_type'
            ]
        elif exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
            aggregation_groupby += [
                'campaign_id', 'campaign_name', 'campaign_type', 'ad_group_id',
                'ad_group_name'
            ]
        return aggregation_groupby



from __future__ import annotations

from googleads_housekeeper import views
from googleads_housekeeper.domain import commands
from googleads_housekeeper.domain import events
from googleads_housekeeper.domain.core import task
from googleads_housekeeper.services import enums


class TestTask:

    def test_create_task_populated_repository(self, bus):
        cmd = commands.SaveTask(exclusion_rule='clicks > 0', customer_ids='1')
        bus.handle(cmd)
        assert bus.uow.tasks.list()

    def test_create_task_publish_task_create_event(self, bus, fake_publisher):
        cmd = commands.SaveTask(exclusion_rule='clicks > 0', customer_ids='1')
        bus.handle(cmd)
        assert isinstance(fake_publisher.events[0], events.TaskCreated)

    def test_create_task_process_queue(self, bus):
        cmd = commands.SaveTask(exclusion_rule='clicks > 0', customer_ids='1')
        bus.handle(cmd)
        assert not bus.queue

    def test_tasks_view_is_not_empty_after_creating_task(self, bus):
        cmd = commands.SaveTask(exclusion_rule='clicks > 0', customer_ids='1')
        bus.handle(cmd)
        assert views.tasks(bus.uow)

    def test_update_task_change_task_in_repository(self, bus):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        update_cmd = commands.SaveTask(exclusion_rule='clicks > 1',
                                       customer_ids='2',
                                       task_id=task_id)
        bus.handle(update_cmd)
        assert bus.uow.tasks.get(task_id).exclusion_rule == 'clicks > 1'
        assert bus.uow.tasks.get(task_id).customer_ids == '2'

    def test_update_task_publish_task_updated_event(self, bus, fake_publisher):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        update_cmd = commands.SaveTask(exclusion_rule='clicks > 1',
                                       customer_ids='2',
                                       task_id=task_id)
        bus.handle(update_cmd)
        assert events.TaskCreated(task_id) in fake_publisher.events
        assert events.TaskUpdated(task_id) in fake_publisher.events

    def test_update_task_with_schedule_publish_task_schedule_updated_event(
            self, bus, fake_publisher):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1',
                                       schedule='1')
        task_id = bus.handle(create_cmd)
        assert events.TaskCreated(task_id) in fake_publisher.events
        assert events.TaskWithScheduleCreated(
            task_id, task_name=None, schedule='1') in fake_publisher.events

    def test_update_task_process_queue(self, bus):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        update_cmd = commands.SaveTask(exclusion_rule='clicks > 1',
                                       customer_ids='2',
                                       task_id=task_id)
        bus.handle(update_cmd)
        assert not bus.queue

    def test_delete_task_makes_task_status_inactive(self, bus):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        delete_cmd = commands.DeleteTask(task_id=task_id)
        bus.handle(delete_cmd)
        assert bus.uow.tasks.get(task_id).status == task.TaskStatus.INACTIVE

    def test_delete_task_publishes_task_deleted_event(self, bus,
                                                      fake_publisher):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        delete_cmd = commands.DeleteTask(task_id=task_id)
        bus.handle(delete_cmd)
        assert events.TaskCreated(task_id) in fake_publisher.events
        assert events.TaskDeleted(task_id) in fake_publisher.events

    def test_delete_task_schedule_publishes_task_schedule_deleted_event(
            self, bus, fake_publisher):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        delete_cmd = commands.DeleteTask(task_id=task_id)
        bus.handle(delete_cmd)
        assert events.TaskScheduleDeleted(task_id) in fake_publisher.events

    def test_delete_task_process_queue(self, bus, fake_publisher):
        create_cmd = commands.SaveTask(exclusion_rule='clicks > 0',
                                       customer_ids='1')
        task_id = bus.handle(create_cmd)
        delete_cmd = commands.DeleteTask(task_id=task_id)
        bus.handle(delete_cmd)
        assert not bus.queue


class TestAllowlisting:

    def test_add_to_allowlisting_create_entry_in_repository(self, bus):
        allowlisting_cmd = commands.AddToAllowlisting(
            type=enums.PlacementTypeEnum.MOBILE_APPLICATION,
            name='fake_name',
            account_id='1')
        bus.handle(allowlisting_cmd)
        assert bus.uow.allowlisting.list()

    def test_remove_from_allowlisting_empties_in_repository(self, bus):
        allowlisting_cmd = commands.AddToAllowlisting(
            type=enums.PlacementTypeEnum.MOBILE_APPLICATION,
            name='fake_name',
            account_id='1')
        bus.handle(allowlisting_cmd)
        removal_allowlisting_cmd = commands.RemoveFromAllowlisting(
            type=enums.PlacementTypeEnum.MOBILE_APPLICATION,
            name='fake_name',
            account_id='1')
        bus.handle(removal_allowlisting_cmd)
        assert not bus.uow.allowlisting.list()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import json
import logging
from dataclasses import asdict

from googleads_housekeeper.domain import events


class BasePublisher:
    ...


class RedisPublisher(BasePublisher):

    def __init__(self,
                 client: 'redis.Redis',
                 topic_prefix: str | None = None) -> None:
        self.client = client
        self.topic_prefix = topic_prefix

    def publish(self, topic: str, event: events.Event) -> None:
        if self.topic_prefix:
            topic = f'{self.topic_prefix}_{topic}'
        self.client.publish(topic, json.dumps(asdict(event), default=str))


class GoogleCloudPubSubPublisher(BasePublisher):

    def __init__(self,
                 client: 'google.cloud.pubsub_v1.PublisherClient',
                 project_id: str,
                 topic_prefix: str | None = None) -> None:
        self.client = client
        self.project_id = project_id
        self.topic_prefix = topic_prefix

    def publish(self, topic: str, event: events.Event) -> None:
        if self.topic_prefix:
            topic = f'{self.topic_prefix}_{topic}'
        topic_name = f'projects/{self.project_id}/topics/{topic}'
        data = str(json.dumps(asdict(event), default=str)).encode('utf-8')
        future = self.client.publish(topic_name, data=data)
        future.result()


class LogPublisher(BasePublisher):

    def __init__(self, topic_prefix: str | None = None) -> None:
        self.topic_prefix = topic_prefix

    def publish(self, topic: str, event: events.Event) -> None:
        if self.topic_prefix:
            topic = f'{self.topic_prefix}_{topic}'
        logging.info("Published to topic '%s': %s", topic, asdict(event))



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains classes and function to work with Placements from Google Ads API."""
from __future__ import annotations

import dataclasses
import datetime
import uuid

import gaarf

from googleads_housekeeper.domain.core import entity
from googleads_housekeeper.services import enums


class Placements(entity.ExcludableEntity):
    """Contains placement meta information and it's performance.

    Placements is a wrapper on BaseQuery that builds GAQL query (located in
    `query_text` attribute) based on provided and validated inputs.

    Attributes:
        query_text: Gaarf query to be sent to Google Ads API.
    """

    _TODAY = datetime.datetime.today()
    _START_DATE = _TODAY - datetime.timedelta(days=7)
    _END_DATE = _TODAY - datetime.timedelta(days=1)
    _CAMPAIGN_TYPES = {c.name for c in enums.CampaignTypeEnum}
    _PLACEMENT_TYPES = {p.name for p in enums.PlacementTypeEnum}
    _NON_EXCLUDABLE_PLACEMENTS = ('youtube.com', 'mail.google.com',
                                  'adsenseformobileapps.com')
    base_query_text = """
        SELECT
            customer.descriptive_name AS account_name,
            customer.id AS customer_id,
            campaign.id AS campaign_id,
            campaign.name AS campaign_name,
            campaign.advertising_channel_type AS campaign_type,
            ad_group.id AS ad_group_id,
            ad_group.name AS ad_group_name,
            {placement_level_granularity}.{parent_url} AS base_url,
            {placement_level_granularity}.target_url AS url,
            {placement_level_granularity}.placement AS placement,
            {placement_level_granularity}.placement_type AS placement_type,
            {placement_level_granularity}.resource_name AS resource_name,
            {placement_level_granularity}.display_name AS name,
            {placement_level_granularity}.resource_name~0 AS criterion_id,
            metrics.clicks AS clicks,
            metrics.impressions AS impressions,
            metrics.cost_micros / 1e6 AS cost,
            metrics.conversions AS conversions,
            metrics.video_views AS video_views,
            metrics.interactions AS interactions,
            metrics.all_conversions AS all_conversions,
            metrics.all_conversions_value AS all_conversions_value,
            metrics.view_through_conversions AS view_through_conversions,
            metrics.conversions_value AS conversions_value
        FROM {placement_level_granularity}
        WHERE segments.date >= "{start_date}"
            AND segments.date <= "{end_date}"
            AND {placement_level_granularity}.placement_type IN
                ("{placement_types}")
            AND {placement_level_granularity}.target_url NOT IN
                ("{non_excludable_placements}")
            AND campaign.advertising_channel_type IN ("{campaign_types}")
            AND metrics.clicks >= {clicks}
            AND metrics.impressions > {impressions}
            AND metrics.ctr < {ctr}
            AND metrics.cost_micros >= {cost}
        """

    def __init__(self,
                 placement_types: tuple[str, ...] | None = None,
                 campaign_types: tuple[str, ...] | None = None,
                 placement_level_granularity: str = 'group_placement_view',
                 start_date: str = _START_DATE.strftime('%Y-%m-%d'),
                 end_date: str = _END_DATE.strftime('%Y-%m-%d'),
                 clicks: int = 0,
                 cost: int = 0,
                 impressions: int = 0,
                 ctr: float = 1.0):
        """Creates Gaarf query for fetching placements data.

        Args:
            placement_types: List of campaign types that need to be fetched.
            placement_types: List of placement types that need to be fetched
                for exclusion.
            placement_level_granularity: API Resource to fetch data from.
            start_date: Start_date of the period.
            end_date: Start_date of the period.
            clicks: Number of clicks for the period.
            impressions: Number of impressions for the period.
            cost: Cost for the period.
            impressions: Impressions for the period.
            ctr: Average CTR for the period.

        Raises:
            ValueError: If campaign_type, placement_type or
            placement_level_granularity are incorrect.
        """
        if campaign_types:
            if isinstance(campaign_types, str):
                campaign_types = tuple(campaign_types.split(','))
            if (wrong_types :=
                    set(campaign_types).difference(self._CAMPAIGN_TYPES)):
                raise ValueError('Wrong campaign type(s): ',
                                 ', '.join(wrong_types))
            self.campaign_types = '","'.join(campaign_types)
        else:
            self.campaign_types = '","'.join(self._CAMPAIGN_TYPES)
        if placement_types:
            if isinstance(placement_types, str):
                placement_types = tuple(placement_types.split(','))
            if (wrong_types :=
                    set(placement_types).difference(self._PLACEMENT_TYPES)):
                raise ValueError('Wrong placement(s): ', ', '.join(wrong_types))
            self.placement_types = '","'.join(placement_types)
        else:
            self.placement_types = '","'.join(self._PLACEMENT_TYPES)

        if placement_level_granularity not in ('detail_placement_view',
                                               'group_placement_view'):
            raise ValueError(
                "Only 'detail_placement_view' or 'group_placement_view' "
                'can be specified!')
        self.placement_level_granularity = placement_level_granularity

        self.validate_dates(start_date, end_date)
        self.start_date = start_date
        self.end_date = end_date
        self.non_excludable_placements = '","'.join(
            self._NON_EXCLUDABLE_PLACEMENTS)
        self.parent_url = ('group_placement_target_url'
                           if self.placement_level_granularity
                           == 'detail_placement_view' else 'target_url')
        self.clicks = clicks
        self.impressions = impressions
        self.cost = int(cost * 1e6)
        self.ctr = ctr
        self.query_text = self.base_query_text.format(**self.__dict__)

    def validate_dates(self, start_date: str, end_date: str) -> None:
        """Checks whether provides start and end dates are valid.

        Args:
            start_date: Date in "YYYY-MM-DD" format.
            end_date: Date in "YYYY-MM-DD" format.

        Raises:
            ValueError: if start or end_date have incorrect format or
            start_date greater than end_date.
        """
        if not self.is_valid_date(start_date):
            raise ValueError(f'Invalid start_date: {start_date}')

        if not self.is_valid_date(end_date):
            raise ValueError(f'Invalid end_date: {end_date}')

        if (datetime.datetime.strptime(start_date, '%Y-%m-%d')
                > datetime.datetime.strptime(end_date, '%Y-%m-%d')):
            raise ValueError('start_date cannot be greater than end_date: '
                             f'{start_date} > {end_date}')

    def is_valid_date(self, date_string: str) -> bool:
        """Validates date.

        Args:
            date_string: Date to be validated.

        Returns:
            Whether or not the date is a string in "YYYY-MM-DD" format.

        Raises:
            ValueError: If string format is incorrect.
        """
        try:
            datetime.datetime.strptime(date_string, '%Y-%m-%d')
            return True
        except ValueError:
            return False


class PlacementsConversionSplit(Placements):
    """Placement conversion performance by each conversion name.

    Attributes:
        query_text: Gaarf query to be sent to Google Ads API.
    """
    _TODAY = datetime.datetime.today()
    _START_DATE = _TODAY - datetime.timedelta(days=7)
    _END_DATE = _TODAY - datetime.timedelta(days=1)

    base_query_text = """
        SELECT
            campaign.advertising_channel_type AS campaign_type,
            ad_group.id AS ad_group_id,
            segments.conversion_action_name AS conversion_name,
            {placement_level_granularity}.placement AS placement,
            metrics.conversions AS conversions,
            metrics.all_conversions AS all_conversions
        FROM {placement_level_granularity}
        WHERE segments.date >= "{start_date}"
            AND segments.date <= "{end_date}"
            AND {placement_level_granularity}.placement_type IN
                ("{placement_types}")
            AND {placement_level_granularity}.target_url NOT IN
                ("{non_excludable_placements}")
            AND campaign.advertising_channel_type IN ("{campaign_types}")
    """

    def __init__(
        self,
        placement_types: tuple[str, ...] | None = None,
        campaign_types: tuple[str, ...] | None = None,
        placement_level_granularity: str = 'group_placement_view',
        start_date: str = _START_DATE.strftime('%Y-%m-%d'),
        end_date: str = _END_DATE.strftime('%Y-%m-%d')
    ) -> None:
        """Creates Gaarf query for fetching placements conversion split data.
        Args:
            placement_types: List of placement types that need to be fetched
                for exclusion.
            campaign_types: List of campaign types that need to be fetched.
            placement_level_granularity: API Resource to fetch data from.
            start_date: Start_date of the period.
            end_date: Start_date of the period.
        """
        super().__init__(placement_types, campaign_types,
                         placement_level_granularity, start_date, end_date)
        self.query_text = self.base_query_text.format(**self.__dict__)


class PlacementExclusionLists(gaarf.base_query.BaseQuery):
    """Contains Gaarf query for getting placement exclusion lists.

    Attributes:
        query_text: Gaarf query to be sent to Google Ads API.
    """

    def __init__(self):
        """Creates Gaarf query for placement exclusion lists."""
        self.query_text = """
            SELECT
                shared_set.name AS name,
                shared_set.resource_name AS resource_name
            FROM shared_set
            WHERE shared_set.type = 'NEGATIVE_PLACEMENTS'
            AND shared_set.status = 'ENABLED'
        """


class AlreadyExcludedPlacements(gaarf.base_query.BaseQuery):
    """Contains Gaarf query for negative placement criteria.

    Negative placements can be specified on multiple levels (ACCOUNT, CAMPAIGN,
    AD_GROUP). AlreadyExcludedPlacements builds the query depending on the
    specified level.

    Attributes:
        query_text: Gaarf query to be sent to Google Ads API.
    """

    _PLACEMENT_TYPES = '","'.join([
        'PLACEMENT', 'YOUTUBE_CHANNEL', 'YOUTUBE_VIDEO', 'MOBILE_APPLICATION',
        'MOBILE_APP_CATEGORY', 'WEBPAGE'
    ])

    base_query_text = """
        SELECT
            {level}.id AS level_id,
            {criterion_resource_name}.placement.url AS website_url,
            {criterion_resource_name}.mobile_application.app_id AS app_id,
            {criterion_resource_name}.youtube_video.video_id AS video_id,
            {criterion_resource_name}.youtube_channel.channel_id AS channel_id
        FROM {criterion_resource_name}
        WHERE {criterion_resource_name}.type IN ("{placement_types}")
    """

    def __init__(self, exclusion_level: enums.ExclusionTypeEnum) -> None:
        """Creates Gaarf query based on specified exclusion level."""
        if exclusion_level == enums.ExclusionLevelEnum.ACCOUNT:
            self.query_text = self.base_query_text.format(
                level='customer',
                criterion_resource_name='customer_negative_criterion',
                placement_types=self._PLACEMENT_TYPES)
        elif exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
            self.query_text = self.base_query_text.format(
                level='campaign',
                criterion_resource_name='campaign_criterion',
                placement_types=self._PLACEMENT_TYPES)
            self.query_text += ' AND campaign_criterion.negative = true'
        elif exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
            self.query_text = self.base_query_text.format(
                level='ad_group',
                criterion_resource_name='ad_group_criterion',
                placement_types=self._PLACEMENT_TYPES)
            self.query_text += ' AND ad_group_criterion.negative = true'


@dataclasses.dataclass
class AllowlistedPlacement:
    """Contains information on allowlisting status of a given placement.

    Attributes:
        type: Placement type (i.e website or YouTube channel).
        name: Name of a placement (i.e. website url, channel id)
        account_id: Account where this placement is allowlisted.
        id: Unique identifier of the allowlisted placement.
    """
    type: enums.PlacementTypeEnum
    name: str
    account_id: str
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))



from __future__ import annotations

import gaarf
import pytest

from googleads_housekeeper.domain.placement_handler import placement_fetcher
from googleads_housekeeper.services import enums

_BASE_PLACEMENT_INFO = [
    'example.com',
    'WEBSITE',
    'example.com',
    '1234',
    'example.com',
]

_BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID = _BASE_PLACEMENT_INFO + [1111]

_BASE_REPORT_COLUMN_NAMES = [
    'placement',
    'placement_type',
    'name',
    'criterion_id',
    'url',
]

_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID = _BASE_REPORT_COLUMN_NAMES + [
    'customer_id'
]


class TestPlacementFetcher:

    @pytest.fixture
    def test_client(self, mocker):
        mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
        return gaarf.api_clients.GoogleAdsApiClient()

    @pytest.fixture
    def fetcher(self, test_client):
        return placement_fetcher.PlacementFetcher(
            gaarf.query_executor.AdsReportFetcher(test_client))

    @pytest.fixture
    def placements(self):
        return gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2222, 3331, 2, 20, 0.2],
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2222, 3332, 5, 50, 0.5],
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2223, 3333, 3, 30, 0.3],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID +
            ['campaign_id', 'ad_group_id', 'clicks', 'impressions', 'cost'])

    def test_aggregate_placements_returns_same_report_for_ad_group_level_aggregation(
            self, fetcher, placements):

        aggregated_report = fetcher._aggregate_placements(
            placements=placements,
            exclusion_level=enums.ExclusionLevelEnum.AD_GROUP,
            perform_relative_aggregations=False)

        assert aggregated_report == placements

    def test_aggregate_placements_returns_correct_report_for_campaign_level_aggregation(
            self, fetcher, placements):

        expected_report = gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID + [2222, 7, 70, 0.7],
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID + [2223, 3, 30, 0.3],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID +
            ['campaign_id', 'clicks', 'impressions', 'cost'])

        aggregated_report = fetcher._aggregate_placements(
            placements=placements,
            exclusion_level=enums.ExclusionLevelEnum.CAMPAIGN,
            perform_relative_aggregations=False)

        assert aggregated_report == expected_report

    def test_aggregate_placements_returns_correct_report_for_account_level_aggregation(
            self, fetcher, placements):

        expected_report = gaarf.report.GaarfReport(
            results=[_BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID + [10, 100, 1.0]],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID +
            ['clicks', 'impressions', 'cost'])

        aggregated_report = fetcher._aggregate_placements(
            placements=placements,
            exclusion_level=enums.ExclusionLevelEnum.ACCOUNT,
            perform_relative_aggregations=False)

        assert aggregated_report == expected_report

    def test_aggregate_placements_with_relative_aggretations_returns_extra_columns_in_report(
            self, fetcher, placements):

        expected_report = gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2222, 7, 70, 0.7, 0.1, 0.1, 10.0],
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2223, 3, 30, 0.3, 0.1, 0.1, 10.0],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID + [
                'campaign_id', 'clicks', 'impressions', 'cost', 'ctr',
                'avg_cpc', 'avg_cpm'
            ])

        aggregated_report = fetcher._aggregate_placements(
            placements=placements,
            exclusion_level=enums.ExclusionLevelEnum.CAMPAIGN,
            perform_relative_aggregations=True)

        assert aggregated_report == expected_report

    def test_join_conversion_split(self, fetcher):
        placements = gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2222, 3331, 2, 20, 0.2],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID +
            ['campaign_id', 'ad_group_id', 'clicks', 'impressions', 'cost'])
        conversion_split_data = gaarf.report.GaarfReport(
            results=[
                [3331, 'example.com', 'test-conversion', 10, 100],
                [3331, 'example.com', 'test-conversion-2', 10, 100],
            ],
            column_names=[
                'ad_group_id', 'placement', 'conversion_name', 'conversions',
                'all_conversions'
            ])
        expected_report = gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                [2222, 3331, 2, 20, 0.2, 'test-conversion', 10, 100],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID + [
                'campaign_id', 'ad_group_id', 'clicks', 'impressions', 'cost',
                'conversion_name', 'conversions_', 'all_conversions_'
            ])
        joined_report = fetcher._join_conversion_split(placements,
                                                       conversion_split_data,
                                                       'test-conversion')
        assert joined_report == expected_report

    def test_aggregate_placements_with_conversion_name_returns_extra_cpa_columns_in_report(
            self, fetcher):

        report = gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO_WITH_CUSTOMER_ID +
                ['test-conversion', 10, 1, 10],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES_WITH_CUSTOMER_ID +
            ['conversion_name', 'cost', 'conversions_', 'all_conversions_'])
        expected_report = gaarf.report.GaarfReport(
            results=[
                _BASE_PLACEMENT_INFO +
                ['test-conversion', 1111, 10, 1, 10, 10.0, 1.0],
            ],
            column_names=_BASE_REPORT_COLUMN_NAMES + [
                'conversion_name', 'customer_id', 'cost', 'conversions_',
                'all_conversions_', 'cost_per_conversion_',
                'cost_per_all_conversion_'
            ])

        aggregated_report = fetcher._aggregate_placements(
            placements=report,
            exclusion_level=enums.ExclusionLevelEnum.ACCOUNT,
            perform_relative_aggregations=True)

        assert aggregated_report == expected_report



from __future__ import annotations

import pytest

from googleads_housekeeper.domain.core import rules_parser


class TestRulesParser:

    @pytest.fixture
    def raw_rules(self):
        return [
            'GOOGLE_ADS_INFO:clicks > 0,cost > 100',
            'GOOGLE_ADS_INFO:placement_type = MOBILE_APPLICATION,ctr = 0',
            'GOOGLE_ADS_INFO:conversions = 0,WEBSITE_INFO:title regexp game'
        ]

    @pytest.fixture
    def implicit_raw_rules(self):
        return [
            'clicks > 0,cost > 100',
            'placement_type = MOBILE_APPLICATION,ctr = 0',
            'conversions = 0,WEBSITE_INFO:title regexp game'
        ]

    @pytest.fixture
    def rules_expression(self):
        return (
            'GOOGLE_ADS_INFO:clicks > 0 AND GOOGLE_ADS_INFO:cost > 100'
            ' OR GOOGLE_ADS_INFO:placement_type = MOBILE_APPLICATION AND '
            'GOOGLE_ADS_INFO:ctr = 0 OR GOOGLE_ADS_INFO:conversions = 0 AND '
            'WEBSITE_INFO:title regexp game')

    @pytest.fixture
    def expected_rules(self):
        return [
            [
                rules_parser.Rule('GOOGLE_ADS_INFO', 'clicks > 0'),
                rules_parser.Rule('GOOGLE_ADS_INFO', 'cost > 100')
            ],
            [
                rules_parser.Rule('GOOGLE_ADS_INFO',
                                  'placement_type = MOBILE_APPLICATION'),
                rules_parser.Rule('GOOGLE_ADS_INFO', 'ctr = 0')
            ],
            [
                rules_parser.Rule('GOOGLE_ADS_INFO', 'conversions = 0'),
                rules_parser.Rule('WEBSITE_INFO', 'title regexp game')
            ],
        ]

    def test_parser_generate_rules_from_raw_rules_with_excplicit_types(
            self, raw_rules, expected_rules):
        rules = rules_parser.generate_rules(raw_rules)
        assert rules == expected_rules

    def test_parser_generate_rules_from_raw_rules_with_implicit_types(
            self, implicit_raw_rules, expected_rules):
        rules = rules_parser.generate_rules(implicit_raw_rules)
        assert rules == expected_rules

    def test_parser_generate_rules_from_expression(self, rules_expression,
                                                   expected_rules):
        rules = rules_parser.generate_rules(rules_expression)
        assert rules == expected_rules



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import os
import inspect
from gaarf.api_clients import BaseClient, GoogleAdsApiClient

if os.getenv("ADS_HOUSEKEEPER_DEPLOYMENT_TYPE") == "Google Cloud":
  try:
    from google.cloud import firestore, pubsub_v1
  except ImportError as e:
    raise ImportError(
        "Please install googleads-housekeeper with Google Cloud support - `pip install googleads-housekeeper[gcp]`"
    ) from e

from googleads_housekeeper.adapters import orm, notifications, publisher
from googleads_housekeeper.services import handlers, messagebus, unit_of_work


class Bootstrapper:

  def __init__(self,
               type: str = "Local",
               topic_prefix: str | None = None) -> None:
    self.notification_service = notifications.ConsoleNotifications()
    if type == "Local":
      self.bootstrapper = LocalBootstrapper(topic_prefix=topic_prefix)
    if type == "Google Cloud":
      self.bootstrapper = GoogleCloudBootstrapper(topic_prefix=topic_prefix)
    if type == "Dev":
      self.bootstrapper = DevBootstrapper(topic_prefix=topic_prefix)

  def bootstrap_app(self,
                    ads_api_client=None,
                    start_orm=False,
                    uow=None,
                    notification_service=None,
                    publish_service=None):
    return self.bootstrapper._bootstrap(
        ads_api_client=ads_api_client,
        start_orm=start_orm,
        uow=uow,
        notification_service=notification_service,
        publish_service=publish_service)

  def _bootstrap(self, ads_api_client, start_orm, uow, notification_service,
                 publish_service):
    return bootstrap(
        ads_api_client=ads_api_client,
        start_orm=start_orm or self.start_orm,
        uow=uow or self.uow,
        notification_service=notification_service or self.notification_service,
        publish_service=publish_service or self.publish_service)


class GoogleCloudBootstrapper(Bootstrapper):

  def __init__(self,
               project_id: str = os.getenv("GOOGLE_CLOUD_PROJECT"),
               start_orm: bool = False,
               topic_prefix: str | None = None) -> None:
    from google.cloud import firestore, pubsub_v1
    self.start_orm = False
    self.uow = unit_of_work.FirestoreUnitOfWork(
        firestore.Client(project=project_id))
    self.notification_service = notifications.GoogleCloudAppEngineEmailNotifications(
        project_id)
    self.publish_service = publisher.GoogleCloudPubSubPublisher(
        client=pubsub_v1.PublisherClient(),
        project_id=project_id,
        topic_prefix=topic_prefix)


class LocalBootstrapper(Bootstrapper):

  def __init__(self, topic_prefix: str | None = None) -> None:
    self.start_orm = True
    self.uow = unit_of_work.SqlAlchemyUnitOfWork(get_db_uri())
    self.publish_service = publisher.LogPublisher(topic_prefix)
    self.notification_service = notifications.ConsoleNotifications()


class DevBootstrapper(Bootstrapper):

  def __init__(self, topic_prefix: str | None = None) -> None:
    self.start_orm = True
    self.uow = unit_of_work.SqlAlchemyUnitOfWork(
        f"sqlite:///{os.getcwd()}/ads_housekeeper_tasks.db")
    self.publish_service = publisher.LogPublisher(topic_prefix)
    self.notification_service = notifications.ConsoleNotifications()


def get_db_uri(host: str = os.environ.get("DB_HOST", "localhost"),
               port: int = int(os.environ.get("DB_PORT", 54321)),
               user: str = os.environ.get("DB_USER", "ads_housekeeper"),
               password: str = os.environ.get("DB_PASSWORD", "ads_housekeeper"),
               db_name: str = os.environ.get("DB_NAME", "ads_housekeeper"),
               db_engine: str = os.environ.get("DB_ENGINE",
                                               "postgresql")) -> str:
  if dev_db := os.environ.get("DEV_DB_HOST"):
    return f"sqlite:///{dev_db}"
  return f"{db_engine}://{user}:{password}@{host}:{port}/{db_name}"


def get_default_google_ads_api_client():
  return GoogleAdsApiClient(
      path_to_config=get_path_to_ads_config(), version=get_ads_api_version())


def get_ads_api_version() -> str:
  return os.environ.get("GOOGLE_ADS_API_VERSION", "v14")


def get_path_to_ads_config() -> str:
  return os.environ.get("GOOGLE_ADS_PATH_TO_CONFIG", "~/google-ads.yaml")


def get_notification_service() -> notifications.BaseNotifications:
  notification_service_factory = notifications.NotificationFactory()
  notification_type = os.environ.get("ADS_HOUSEKEEPER_NOTIFICATION_TYPE",
                                     "console")
  notification_service = notification_service_factory.create_nofication_service(
      notification_type=notification_type)
  return notification_service


def bootstrap(
    ads_api_client: BaseClient | None = None,
    start_orm: bool = True,
    uow: unit_of_work.AbstractUnitOfWork = unit_of_work.SqlAlchemyUnitOfWork(
        get_db_uri()),
    notification_service: notifications
    .BaseNotifications = get_notification_service(),
    publish_service: publisher.BasePublisher = publisher.LogPublisher()
) -> messagebus.MessageBus:

  if start_orm:
    orm.start_mappers(engine=uow.engine)

  dependencies = {
      "uow": uow,
      "ads_api_client": ads_api_client or get_default_google_ads_api_client(),
      "notification_service": notification_service,
      "publisher": publish_service
  }
  injected_event_handlers = {
      event_type: [
          inject_dependencies(handler, dependencies) for handler in handlers
      ] for event_type, handlers in handlers.EVENT_HANDLERS.items()
  }
  injected_command_handlers = {
      command_type: inject_dependencies(handler, dependencies)
      for command_type, handler in handlers.COMMAND_HANDLERS.items()
  }
  return messagebus.MessageBus(
      uow=uow,
      event_handlers=injected_event_handlers,
      command_handlers=injected_command_handlers,
      dependencies=dependencies)


def inject_dependencies(handler, dependencies):
  params = inspect.signature(handler).parameters
  deps = {
      name: dependency
      for name, dependency in dependencies.items()
      if name in params
  }
  return lambda message: handler(message, **deps)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc
from dataclasses import asdict
from typing import Any

from google.cloud.firestore_v1.base_query import FieldFilter

from googleads_housekeeper.domain.core import task


class AbstractRepository(abc.ABC):

    def __init__(self):
        self.seen = set()

    def add(self, task: task.Task):
        self._add(task)

    def delete(self, task_id: str):
        self._delete(task_id)

    def get(self, task_id) -> task.Task:
        return self._get(task_id)

    def get_by_conditions(self, conditions: dict[str, Any]) -> task.Task:
        return self._get_by_conditions(conditions)

    def get_by_condition(self, condition_name: str,
                         condition_value: str) -> task.Task:
        return self._get_by_conditions({condition_name: condition_value})

    def list(self) -> list[task.Task]:
        return self._list()

    def update(self, task_id: int, update_dict: dict[str, str]) -> task.Task:
        return self._update(task_id, update_dict)

    @abc.abstractmethod
    def _add(self, task: task.Task):
        ...

    @abc.abstractmethod
    def _get(self, task_id) -> task.Task:
        ...

    @abc.abstractmethod
    def _get_by_conditions(self, conditions: dict[str, Any]) -> task.Task:
        ...

    @abc.abstractmethod
    def _list(self) -> list[task.Task]:
        ...

    @abc.abstractmethod
    def _update(self, task_id, update_dict):
        ...

    @abc.abstractmethod
    def _delete(self, task_id):
        ...


class SqlAlchemyRepository(AbstractRepository):

    def __init__(self, session, entity=task.Task):
        super().__init__()
        self.session = session
        self.entity = entity

    def _add(self, task):
        self.session.add(task)

    def _get(self, task_id):
        return self.session.query(self.entity).filter_by(id=task_id).first()

    def _get_by_conditions(self, conditions: dict[str, Any]):
        query = self.session.query(self.entity)
        for condition_name, condition_value in conditions.items():
            query = query.filter(
                getattr(self.entity, condition_name) == condition_value)
        return query.all()

    def _list(self):
        return self.session.query(self.entity).all()

    def _update(self, task_id, update_dict):
        return self.session.query(
            self.entity).filter_by(id=task_id).update(update_dict)

    def _delete(self, task_id):
        return self.session.query(self.entity).filter_by(id=task_id).delete()


class FirestoreRepository(AbstractRepository):

    def __init__(self, client, entity=task.Task):
        super().__init__()
        self.client = client
        self.entity = entity
        self.collection_name = entity.__name__

    def _add(self, task):
        if hasattr(task, 'id'):
            element_id = task.id
        else:
            element_id = task._id
        element_dict = {}
        for key, value in asdict(task).items():
            if hasattr(value, 'name'):
                value = value.name
            element_dict[key] = value
        self.client.collection(self.collection_name).document(
            str(element_id)).set(element_dict)

    def _get(self, task_id: str):
        doc = self.client.collection(
            self.collection_name).document(task_id).get()
        if doc.exists:
            return self.entity(**doc.to_dict())
        return None

    def _get_by_conditions(self, conditions: dict[str, Any]):
        try:
            query = self.client.collection(self.collection_name)
            for condition_name, condition_value in conditions.items():
                query = query.where(
                    filter=FieldFilter(condition_name, '==', condition_value))
            results = [
                self.entity(**result.to_dict()) for result in query.stream()
            ]
            return results
        except Exception:
            return []

    def _list(self):
        results = self.client.collection(self.collection_name).stream()
        entities = [self.entity(**result.to_dict()) for result in results]
        return entities

    def _update(self, task_id: str, update_dict: dict[str, Any]):
        if doc := self.client.collection(
                self.collection_name).document(task_id):
            doc.update(update_dict)

    def _delete(self, task_id: str) -> None:
        self.client.collection(self.collection_name).document(task_id).delete()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc

import sqlalchemy

from googleads_housekeeper.adapters import repository
from googleads_housekeeper.domain import events
from googleads_housekeeper.domain.core import execution
from googleads_housekeeper.domain.core import settings
from googleads_housekeeper.domain.external_parsers import website_parser
from googleads_housekeeper.domain.external_parsers import youtube_data_parser
from googleads_housekeeper.domain.placement_handler import entities as placement_entities


class AbstractUnitOfWork(abc.ABC):
    tasks: repository.AbstractRepository
    settings: repository.AbstractRepository
    customer_ids: repository.AbstractRepository
    mcc_ids: repository.AbstractRepository
    website_info: repository.AbstractRepository
    youtube_channel_info: repository.AbstractRepository
    youtube_video_info: repository.AbstractRepository
    allowlisting: repository.AbstractRepository
    executions: repository.AbstractRepository
    execution_details: repository.AbstractRepository
    published_events: list[events.Event] = []

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.rollback()

    def commit(self):
        self._commit()

    def collect_new_events(self):
        while self.published_events:
            yield self.published_events.pop(0)

    @abc.abstractmethod
    def _commit(self):
        ...

    @abc.abstractmethod
    def rollback(self):
        ...


class SqlAlchemyUnitOfWork(AbstractUnitOfWork):

    def __init__(self, session_factory: str) -> None:
        self.engine = sqlalchemy.create_engine(session_factory)
        self.session_factory = sqlalchemy.orm.sessionmaker(self.engine)
        self.published_events: list[events.Event] = []

    def __enter__(self) -> None:
        self.session = self.session_factory()
        self.tasks = repository.SqlAlchemyRepository(self.session)
        self.settings = repository.SqlAlchemyRepository(
            session=self.session, entity=settings.Config)
        self.customer_ids = repository.SqlAlchemyRepository(
            session=self.session, entity=settings.CustomerIds)
        self.mcc_ids = repository.SqlAlchemyRepository(
            session=self.session, entity=settings.MccIds)
        self.website_info = repository.SqlAlchemyRepository(
            self.session, entity=website_parser.WebsiteInfo)
        self.youtube_channel_info = repository.SqlAlchemyRepository(
            self.session, entity=youtube_data_parser.ChannelInfo)
        self.youtube_video_info = repository.SqlAlchemyRepository(
            self.session, entity=youtube_data_parser.VideoInfo)
        self.allowlisting = repository.SqlAlchemyRepository(
            self.session, entity=placement_entities.AllowlistedPlacement)
        self.executions = repository.SqlAlchemyRepository(
            self.session, entity=execution.Execution)
        self.execution_details = repository.SqlAlchemyRepository(
            self.session, entity=execution.ExecutionDetails)
        return super().__enter__()

    def __exit__(self, *args):
        super().__exit__(*args)
        self.session.close()

    def _commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()


class FirestoreUnitOfWork(AbstractUnitOfWork):

    def __init__(self, client: 'google.cloud.firestore.Client') -> None:
        self.client = client
        self.published_events: list[events.Event] = []

    def __enter__(self):
        self.tasks = repository.FirestoreRepository(self.client)
        self.settings = repository.FirestoreRepository(
            client=self.client, entity=settings.Config)
        self.customer_ids = repository.FirestoreRepository(
            client=self.client, entity=settings.CustomerIds)
        self.mcc_ids = repository.FirestoreRepository(
            client=self.client, entity=settings.MccIds)
        self.website_info = repository.FirestoreRepository(
            self.client, entity=website_parser.WebsiteInfo)
        self.youtube_channel_info = repository.FirestoreRepository(
            self.client, entity=youtube_data_parser.ChannelInfo)
        self.youtube_video_info = repository.FirestoreRepository(
            self.client, entity=youtube_data_parser.VideoInfo)
        self.allowlisting = repository.FirestoreRepository(
            self.client, entity=placement_entities.AllowlistedPlacement)
        self.executions = repository.FirestoreRepository(
            self.client, entity=execution.Execution)
        self.execution_details = repository.FirestoreRepository(
            self.client, entity=execution.ExecutionDetails)
        return super().__enter__()

    def _commit(self):
        ...

    def rollback(self):
        ...



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining various notification channels."""
from __future__ import annotations

import abc
import dataclasses
import os

import gaarf
import slack_sdk
from google.appengine.api import mail


@dataclasses.dataclass
class MessagePayload:
    """Contains information to be sent via a given notification channel.

    Attributes:
        task_id:
            task identifier.
        task_name:
            task name.
        task_formula:
            task full exclusion_rule.
        placements_excluded_sample:
            optional report with top N placements that were excluded.
        total_placement_excluded:
            number of all excluded placements by a given task.
        recipient:
            who should receive the notification (could be email, slack channel).
    """
    task_id: str
    task_name: str
    task_formula: str
    placements_excluded_sample: gaarf.report.GaarfReport | None
    total_placement_excluded: int
    recipient: str


class BaseNotifications(abc.ABC):
    """Base class for sending notifications."""

    @abc.abstractmethod
    def send(self, payload: MessagePayload) -> None:
        """Transforms payload and sends it via a given notification channel."""
        raise NotImplementedError


class GoogleCloudAppEngineEmailNotifications(BaseNotifications):
    """Implements sending emails with App Engine.

    Attributes:
        project_id: Google Cloud project id.
    """

    def __init__(self, project_id: str) -> None:
        """Initializes GoogleCloudAppEngineEmailNotifications class.

        Args:
            project_id: Id of Google Cloud Project.
        """
        self.project_id = project_id

    def send(self, payload: MessagePayload, **kwargs: str) -> None:
        """Sends payload to via appengine mailer."""
        message = self._prepare_message(payload)
        message.send()

    def _prepare_message(self, payload: MessagePayload) -> mail.Message:
        """Generates message from a payload."""
        task_name = payload.get('task_name', '')
        sender = f'Placements Excluder ({task_name})'
        if placements_excluded := payload.placements_excluded_sample:
            subject = f'{payload.total_placement_excluded} placements were excluded'
            body = self._generate_notification_html(
                payload.task_id, task_name, payload.task_formula, placements_excluded)
        else:
            subject = 'No placements were excluded'
            body = ''
        recipient = payload.recipient.split(',')
        if body:
            return mail.EmailMessage(
                sender=sender, to=recipient, subject=subject, html=body)
        return mail.EmailMessage(
            sender=sender, to=recipient, subject=subject, body=body)

    def _generate_notification_html(self, task_id: str, task_name: str, task_formula: str, placements: gaarf.report.GaarfReport) -> str:
        html_body = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>CPR Placement Exclusion Alert</title>
            <style type="text/css">
        body {
            font-family: Arial, sans-serif;
            margin: 0;
            padding: 20px;
            background-color: #f8f8f8; /* Light background for visual separation */
        }

        .container {
            background-color: #fff; /* White background for content */
            border: 1px solid #ddd;
            padding: 20px;
            border-radius: 8px;
            box-shadow: 0 2px 4px rgba(0,0,0,0.1); /* Subtle shadow for depth */
        }

        h2 {
            color: #2c3e50; /* A darker heading color */
            margin-bottom: 15px;
            font-weight: bold;
        }

        .task-link {
            color: #007bff;
            text-decoration: none;
            font-weight: 600; /* Slightly bolder task link */
        }

        .placement-list {
            list-style: disc;
            margin-left: 20px;
        }

        .logo-placeholder {
            text-align: center;
            margin-top: 20px;
        }

        .logo-placeholder img {
        max-width: 150px; /* Limit logo size */
        }

            </style>
        </head>
        <body>
            <p>Please do not reply to this email.</p>

            <div class="container">
                <h2>CPR Placement Exclusion Alert</h2>"""

        html_body += f"""
                <p>The CPR tool has excluded the following placements
                based on your defined thresholds:</p>
                <p><strong>Task Name: </strong>{task_name}</p>
                <p><strong>Task Logic: </strong>{self._to_human_readable(task_formula)}
                (<a href="https://{self.project_id}.uc.r.appspot.com/newtask?task={task_id}"
                class="task-link"></a>)
                </p>"""

        for placement in placements:
            html_body += f"""
                <ul class="placement-list">
                    <li title={placement.reason}><strong>Placement Name:</strong>
                    {placement["name"]} (<a href="[Placement Link]">
                    {placement["placement"]}</a>)
                    <br/>
                    Reason: {self._to_human_readable(placement.reason)}</li>
                </ul>
            """
        html_body += """
            </div>

            <div class="logo-placeholder">
                <img src="/googleads_housekeeper/rsc/gtech_logo_emblem_rgb_White.png"
                alt="gPS Logo">
            </div>

            <p>Please do not reply to this email.</p>
        </body>
        </html>
        """
        return html_body

    @staticmethod
    def _to_human_readable(text: str) -> str:
        return text.lower().replace('_', ' ')


class SlackNotifications(BaseNotifications):
    """Implements sending emails with Slack.

    Attributes:
        bot_token: Slack API bot token.
        client: initialized Slack WebClient.
    """

    def __init__(self, bot_token: str) -> None:
        """Initializes SlackNotifications class.

        Args:
            bot_token: Slack API Token.
        """
        self.client = slack_sdk.WebClient(token=bot_token)

    def send(self, payload: MessagePayload) -> None:
        """Sends payload to Slack channel."""
        task_name = payload.task_name
        channel = payload.recipient
        if not (placements_excluded_sample :=
                payload.placements_excluded_sample):
            self.client.chat_postMessage(
                channel=channel,
                text=f'{task_name}: No placements were excluded')
        else:
            file = placements_excluded_sample['name'].to_pandas().to_csv(
                index=False, sep='\t')
            self.client.files_upload(
                channels=channel,
                initial_comment=(
                    f'{task_name}: '
                    f'{payload.total_placement_excluded} placements excluded'),
                filename=f'{task_name}.tsv' if task_name else 'cpr.tsv',
                content=file)


class ConsoleNotifications(BaseNotifications):
    """Implements sending notification to standard output."""

    def send(self, payload: MessagePayload) -> None:
        """Prints payload to standard output."""
        print(payload)


class NullNotifications(BaseNotifications):
    """Implements special case for incorrect notification channel."""

    def __init__(self, notification_type: str) -> None:
        self.notification_type = notification_type

    def send(self, payload: MessagePayload) -> None:
        raise ValueError(f'Cannot send notification via '
                         f'"{self.notification_type}" notification channel')


class NotificationFactory:
    """Builds notification service for a given notification type.

    NotificationFactory contains all possible types of notification channels
    defined in the module.

    Attributes:
        types: mapping between name of services and its class.

    """
    types: dict[str, dict[str, BaseNotifications | dict]] = {}

    def __init__(self) -> None:
        self.load_types()

    def load_types(self) -> None:
        self.types['email'] = {
            'type': GoogleCloudAppEngineEmailNotifications,
            'args': {
                'project_id': os.environ.get('GOOGLE_CLOUD_PROJECT')
            }
        }
        self.types['slack'] = {
            'type': SlackNotifications,
            'args': {
                'bot_token': os.environ.get('CPR_SLACK_BOT_TOKEN')
            }
        }
        self.types['console'] = {'type': ConsoleNotifications, 'args': {}}

    def create_nofication_service(
            self, notification_type: str) -> type[BaseNotifications]:
        if notification_type in self.types:
            if args := self.types[notification_type].get('args'):
                return self.types[notification_type].get('type')(**args)
            return self.types[notification_type].get('type')()
        return NullNotifications(notification_type)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field

import pytest
from gaarf.report import GaarfReport

from googleads_housekeeper.domain.core import exclusion_specification
from googleads_housekeeper.domain.core import rules_parser
from googleads_housekeeper.domain.external_parsers import website_parser
from googleads_housekeeper.domain.external_parsers import youtube_data_parser


@dataclass
class FakePlacement:
    campaign_id: int = 1
    campaign_name: str = '01_test_campaign'
    placement: str = 'example.com'
    clicks: int = 10
    ctr: float = 0.4
    placement_type: str = 'WEBSITE'
    extra_info: dict = field(
        default_factory=lambda: {
            'website_info':
                website_parser.WebsiteInfo(
                    placement='example.com', title='example')
        })


@pytest.fixture
def placement():
    return FakePlacement()


class TestAdsExclusionSpecificationEntry:

    @pytest.mark.parametrize('expression', ['single_name', 'single_name >'])
    def test_invalid_expression_length_raises_value_error(self, expression):
        with pytest.raises(ValueError, match='Incorrect expression *'):
            exclusion_specification.AdsExclusionSpecificationEntry(expression)

    def test_invalid_expression_operator_raises_value_error(self):
        with pytest.raises(ValueError, match='Incorrect operator *'):
            exclusion_specification.AdsExclusionSpecificationEntry('clicks ? 0')

    @pytest.fixture(params=[
        'clicks > 1', 'clicks = 10', 'ctr < 0.6', 'placement_type = WEBSITE',
        'placement_type contains WEB', 'campaign_name regexp ^01.+'
    ])
    def ads_exclusion_specification_success(self, request):
        return exclusion_specification.AdsExclusionSpecificationEntry(
            request.param)

    @pytest.fixture(params=[
        'clicks > 100', 'clicks != 10', 'ctr > 0.6',
        'placement_type = MOBILE_APPLICATION', 'placement_type contains MOBILE',
        'campaign_name regexp ^02.+'
    ])
    def ads_exclusion_specification_fail(self, request):
        return exclusion_specification.AdsExclusionSpecificationEntry(
            request.param)

    def test_placement_safisties_ads_exclusion_specification(
            self, placement, ads_exclusion_specification_success):
        result = ads_exclusion_specification_success.is_satisfied_by(placement)
        assert result

    def test_placement_does_not_satisfy_ads_exclusion_specification(
            self, placement, ads_exclusion_specification_fail):
        result = ads_exclusion_specification_fail.is_satisfied_by(placement)
        assert not result[0]

    def test_is_satisfied_by_raises_value_error_when_non_existing_entity_name_is_provided(
            self, placement):
        specification = exclusion_specification.AdsExclusionSpecificationEntry(
            'fake_name > 0')
        with pytest.raises(ValueError):
            specification.is_satisfied_by(placement)


class TestContentExclusionSpecifition:

    @pytest.fixture(params=[
        'title contains games', 'description contains fun',
        'keywords regexp free'
    ])
    def content_exclusion_specification_success(self, request):
        return exclusion_specification.ContentExclusionSpecificationEntry(
            expression=request.param)

    @pytest.fixture(params=[
        'title contains football', 'description contains gloomy',
        'keywords regexp paid'
    ])
    def content_exclusion_specification_failure(self, request):
        return exclusion_specification.ContentExclusionSpecificationEntry(
            expression=request.param)

    def test_website_satisfies_content_exclusion_specification(
            self, placement, content_exclusion_specification_success):
        result = content_exclusion_specification_success.is_satisfied_by(
            placement)
        assert result

    def test_website_does_not_satisfy_content_exclusion_specification(
            self, placement, content_exclusion_specification_failure):
        result = content_exclusion_specification_failure.is_satisfied_by(
            placement)
        assert not result[0]


class TestYouTubeChannelExclusionSpecifition:

    @pytest.fixture
    def youtube_channel_placement(self):
        return FakePlacement(
            placement='1kdjf0skdjfw0dsdf', placement_type='YOUTUBE_CHANNEL')

    @pytest.fixture(params=[
        'title regexp garden*', 'description contains game', 'country = US',
        'viewCount > 10', 'subscriberCount > 1', 'videoCount > 1',
        'topicCategories contains game'
    ])
    def youtube_channel_exclusion_specification_success(self, request):
        return exclusion_specification.YouTubeChannelExclusionSpecificationEntry(
            expression=request.param)

    @pytest.fixture(params=[
        'title regexp football*', 'description contains tv', 'country = GB',
        'viewCount > 100000', 'subscriberCount > 100000', 'videoCount > 100000',
        'topicCategories contains football'
    ])
    def youtube_channel_exclusion_specification_failure(self, request):
        return exclusion_specification.YouTubeChannelExclusionSpecificationEntry(
            expression=request.param)

    def test_youtube_channel_satisfies_youtube_channel_exclusion_specification(
            self, youtube_channel_placement,
            youtube_channel_exclusion_specification_success):
        result = youtube_channel_exclusion_specification_success.is_satisfied_by(
            youtube_channel_placement)
        assert result

    def test_youtube_channel_does_not_satisfy_youtube_channel_exclusion_specification(
            self, youtube_channel_placement,
            youtube_channel_exclusion_specification_failure):
        result = youtube_channel_exclusion_specification_failure.is_satisfied_by(
            youtube_channel_placement)
        assert not result[0]


class TestYouTubeVideoExclusionSpecifition:

    @pytest.fixture
    def youtube_video_placement(self):
        return FakePlacement(placement='jojoh', placement_type='YOUTUBE_VIDEO')

    @pytest.fixture(params=[
        'title regexp garden*', 'description contains game',
        'defaultLanguage = en', 'defaultAudioLanguage = en', 'commentCount > 1',
        'favouriteCount > 1', 'likeCount > 1', 'viewCount > 10',
        'madeForKids = True', 'topicCategories contains game',
        'tags contains garden'
    ])
    def youtube_video_exclusion_specification_success(self, request):
        return exclusion_specification.YouTubeVideoExclusionSpecificationEntry(
            expression=request.param)

    @pytest.fixture(params=[
        'title regexp football*', 'description contains football',
        'defaultLanguage = es', 'defaultAudioLanguage = es',
        'commentCount > 10000', 'favouriteCount > 10000', 'likeCount > 10000',
        'viewCount > 1000000', 'madeForKids = False',
        'topicCategories contains football', 'tags contains football'
    ])
    def youtube_video_exclusion_specification_failure(self, request):
        return exclusion_specification.YouTubeVideoExclusionSpecificationEntry(
            expression=request.param)

    def test_youtube_video_satisfies_youtube_video_exclusion_specification(
            self, youtube_video_placement,
            youtube_video_exclusion_specification_success):
        result = youtube_video_exclusion_specification_success.is_satisfied_by(
            youtube_video_placement)
        assert result

    def test_youtube_video_does_not_satisfy_youtube_video_exclusion_specification(
            self, youtube_video_placement,
            youtube_video_exclusion_specification_failure):
        result = youtube_video_exclusion_specification_failure.is_satisfied_by(
            youtube_video_placement)
        assert not result[0]


class TestExclusionSpecificationEntry:

    @pytest.fixture
    def rules(self):
        return rules_parser.generate_rules(raw_rules=[
            ('GOOGLE_ADS_INFO:clicks > 1, '
             'GOOGLE_ADS_INFO:conversion_name regexp test_conversion'),
            'YOUTUBE_VIDEO_INFO:likeCount > 1',
        ])

    @pytest.fixture
    def ads_specification_entry(self):
        return exclusion_specification.AdsExclusionSpecificationEntry(
            expression='clicks > 1')

    @pytest.fixture
    def ads_specification_entry_conversion_split(self):
        return exclusion_specification.AdsExclusionSpecificationEntry(
            expression='conversion_name regexp test_conversion')

    @pytest.fixture
    def non_ads_specification_entry(self):
        return exclusion_specification.YouTubeVideoExclusionSpecificationEntry(
            expression='likeCount > 1')

    @pytest.fixture
    def sample_exclusion_specification(
            self, ads_specification_entry, non_ads_specification_entry,
            ads_specification_entry_conversion_split):
        return exclusion_specification.ExclusionSpecification(specifications=[[
            ads_specification_entry, ads_specification_entry_conversion_split
        ], [non_ads_specification_entry]])

    @pytest.fixture
    def placements(self):
        return GaarfReport(
            results=[
                [
                    'youtube_video', 'YOUTUBE_VIDEO', 10, 'test_conversion', 0,
                    {
                        'youtube_video_info':
                            youtube_data_parser.VideoInfo(
                                placement='youtube_video', likeCount=10)
                    }
                ],
                ['website', 'WEBSITE', 0, 'real_conversion', 0, {}],
            ],
            column_names=[
                'placement', 'placement_type', 'clicks', 'conversion_name',
                'conversions_', 'extra_info'
            ])

    def test_true_exclusion_specification(self, sample_exclusion_specification):
        assert bool(sample_exclusion_specification) is True

    def test_placement_satisfies_ads_exclusion_specifications_list(
            self, ads_specification_entry):
        """
        If placement satisfies all exclusion specifications in the list,
        it satisfies the whole list.
        """
        specification = exclusion_specification.ExclusionSpecification(
            specifications=[[ads_specification_entry]])
        sample_placement = FakePlacement(clicks=10)
        result = specification.satisfies(sample_placement)
        assert result == ([[
            'GOOGLE_ADS_INFO:clicks > 1',
        ]], {})

    def test_placement_does_not_satisfy_ads_exclusion_specifications_list(
            self, ads_specification_entry):
        """
        If placement does not satisfy at least one exclusion specifications in
        the list, it does not satisfy the whole list.
        """
        specification = exclusion_specification.ExclusionSpecification(
            specifications=[[ads_specification_entry]])
        sample_placement = FakePlacement(clicks=0)
        result = specification.satisfies(sample_placement)
        assert result == ([], {})

    def test_get_correct_ads_entries(self, sample_exclusion_specification,
                                     ads_specification_entry,
                                     ads_specification_entry_conversion_split):
        expected_exclusion_specifications = (
            exclusion_specification.ExclusionSpecification(specifications=[[
                ads_specification_entry,
                ads_specification_entry_conversion_split
            ]]))
        ads_specs = sample_exclusion_specification.ads_specs_entries
        assert ads_specs == expected_exclusion_specifications

    def test_get_correct_non_ads_entries(self, sample_exclusion_specification,
                                         non_ads_specification_entry):
        expected_exclusion_specifications = (
            exclusion_specification.ExclusionSpecification(
                specifications=[[non_ads_specification_entry]]))
        non_ads_specs = sample_exclusion_specification.non_ads_specs_entries
        assert non_ads_specs == expected_exclusion_specifications

    def test_parser_generate_rules_explicit_types(
            self, rules, sample_exclusion_specification):
        specification = exclusion_specification.ExclusionSpecification.from_rules(
            rules)
        assert specification == sample_exclusion_specification

    def test_generate_runtime_options(self, sample_exclusion_specification,
                                      ads_specification_entry_conversion_split):
        runtime_options = sample_exclusion_specification.define_runtime_options(
        )
        assert runtime_options == exclusion_specification.RuntimeOptions(
            is_regular_query=True,
            is_conversion_query=True,
            conversion_name='test_conversion',
            conversion_rules=[ads_specification_entry_conversion_split])

    def test_apply_specifications(self, sample_exclusion_specification,
                                  placements):
        expected_result = GaarfReport(
            results=[
                [
                    'youtube_video', 'YOUTUBE_VIDEO', 10, 'test_conversion', 0,
                    {
                        'youtube_video_info':
                            youtube_data_parser.VideoInfo(
                                placement='youtube_video', likeCount=10)
                    }
                ],
            ],
            column_names=[
                'placement', 'placement_type', 'clicks', 'conversion_name',
                'conversions_', 'extra_info'
            ])
        result = sample_exclusion_specification.apply_specifications(placements)
        assert result['placement'] == expected_result['placement']



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extracts extra info on placements from DB."""
from __future__ import annotations

import gaarf

from googleads_housekeeper.adapters import repository
from googleads_housekeeper.domain.external_parsers import base_parser
from googleads_housekeeper.services import unit_of_work


class PlacementInfoExtractor:
    """Responsible for getting parsed placement information from DB.

    Attributes:
        website_info: Repository containing parsed website data.
        youtube_channel_info: Repository containing parsed YouTube channel data.
        youtube_video_info: Repository containing parsed YouTube video data.
    """

    def __init__(self, uow: unit_of_work.AbstractUnitOfWork) -> None:
        """Creates PlacementInfoExtractor repositories.

        Args:
            uow: An instance of Unit of Work.
        """
        self.website_info = uow.website_info
        self.youtube_channel_info = uow.youtube_channel_info
        self.youtube_video_info = uow.youtube_video_info

    def extract_placement_info(
        self, placement_info: gaarf.report.GaarfRow
    ) -> dict[str, base_parser.EntityInfo]:
        """Fetches parsed information for a given placement.

        Args:
            placement_info:
                A row from report containing placement info from Google Ads.
        Returns:
            Mapping between placement parsed into type (i.e. 'website_info')
            and actual information fetched from DB.
        """
        if placement_info.placement_type == 'WEBSITE':
            return {
                'website_info':
                    self._get_placement_from_repo(self.website_info,
                                                  placement_info.placement)
            }
        if placement_info.placement_type == 'YOUTUBE_CHANNEL':
            return {
                'youtube_channel_info':
                    self._get_placement_from_repo(self.youtube_channel_info,
                                                  placement_info.placement)
            }

        if placement_info.placement_type == 'YOUTUBE_VIDEO':
            return {
                'youtube_video_info':
                    self._get_placement_from_repo(self.youtube_video_info,
                                                  placement_info.placement)
            }
        return {}

    def _get_placement_from_repo(
            self, repo: repository.AbstractRepository,
            placement: str) -> base_parser.EntityInfo | dict:
        """Helper method for getting data from repository.

        Args:
            repository: Repository to get data from.
            placement: Placement identifier (website_url, channel_id, video_id).
        Returns:
            Parsed placement info.
        """
        if parsed_placement_info := repo.get_by_condition(
                'placement', placement):
            return parsed_placement_info[0]
        return {}



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for converting legacy task v1 to v2."""
from __future__ import annotations

from googleads_housekeeper.domain.core import task
from googleads_housekeeper.services import enums


class TaskAdapter:
    """Maps a V1 Task (dict) into a V2 Task object.

    Attributes:
        task_v1_data: A dictionary containing data of a V1 Task object.
    """

    def __init__(self, task_v1_data: dict) -> None:
        """
        Initializes a TaskAdapter object.

        Args:
            task_v1_data:
                A dictionary containing data of a V1 Task object.
        """
        self.task_v1_data = task_v1_data

    def from_task_v1(self) -> task.Task:
        """Converts a V1 Task into a V2 Task object.

        Returns:
            Task: A V2 Task object.
        """
        task_dict = {}

        # Dimensions:
        task_dict['name'] = self.task_v1_data['task_name']
        task_dict['customer_ids'] = self.task_v1_data['customer_id']
        task_dict['date_range'] = self.task_v1_data['lookback_days']
        task_dict['from_days_ago'] = self.task_v1_data['from_days_ago']
        task_dict['exclusion_level'] = enums.ExclusionLevelEnum.AD_GROUP.value
        task_dict['placement_types'] = self._build_placement_types()
        task_dict['status'] = task.TaskStatus.ACTIVE
        task_dict['creation_date'] = None

        task_dict['exclusion_rule'] = self._build_condition_rules()
        task_dict['task_output'] = self._build_task_output()
        return task.Task(**task_dict)

    def _build_placement_types(self) -> str:
        placement_types = set()
        if self.task_v1_data.get('gads_data_youtube'):
            placement_types.add('YOUTUBE_VIDEO')
            placement_types.add('YOUTUBE_CHANNEL')
        if self.task_v1_data.get('gads_data_display'):
            placement_types.add('WEBSITE')
        return ', '.join(sorted(placement_types))

    def _build_condition_rules(self) -> str:
        youtube_filters = self._build_youtube_filters()
        google_ads_filters = self._build_gads_filters()
        return ' AND '.join([google_ads_filters] + youtube_filters)

    def _build_youtube_filters(self) -> list[str]:
        youtube_filters: list[str] = []
        yt_subscriber_operator = self.task_v1_data.get('yt_subscriber_operator')
        yt_subscriber_value = self.task_v1_data.get('yt_subscriber_value')
        yt_view_operator = self.task_v1_data.get('yt_view_operator')
        yt_view_value = self.task_v1_data.get('yt_view_value')
        yt_video_operator = self.task_v1_data.get('yt_video_operator')
        yt_video_value = self.task_v1_data.get('yt_video_value')
        yt_country_operator = self.task_v1_data.get('yt_country_operator')
        yt_country_value = self.task_v1_data.get('yt_country_value')
        yt_language_operator = self.task_v1_data.get('yt_language_operator')
        yt_language_value = self.task_v1_data.get('yt_language_value')
        if (yt_subscriber_operator is not None and
                yt_subscriber_value is not None):
            youtube_filters.append(
                'YOUTUBE_CHANNEL_INFO:subscriberCount '
                f'{yt_subscriber_operator}, {yt_subscriber_value},')
        if yt_view_operator is not None and yt_view_value is not None:
            youtube_filters.append('YOUTUBE_CHANNEL_INFO:yt_view_operator '
                                   f'{yt_view_operator} {yt_view_value}')
        if yt_video_operator is not None and yt_video_value is not None:
            youtube_filters.append('YOUTUBE_CHANNEL_INFO:yt_video_operator '
                                   f'{yt_video_operator} {yt_video_value}')
        if yt_country_operator is not None and yt_country_value is not None:
            youtube_filters.append('YOUTUBE_CHANNEL_INFO:country '
                                   f'{yt_country_operator} {yt_country_value}')
        if yt_language_operator is not None and yt_language_value is not None:
            youtube_filters.append(
                'YOUTUBE_CHANNEL_INFO:defaultLanguage '
                f'{yt_language_operator}, {yt_language_value},')
        has_latin = ('YOUTUBE_CHANNEL_INFO:title has_latin_letters '
                     "{str(self.task_v1_data.get('yt_std_character', False))}")
        youtube_filters.append(has_latin)
        return youtube_filters

    def _build_gads_filters(self) -> str:
        if google_ads_filters := self.task_v1_data.get('gads_filter'):
            return google_ads_filters.replace('metrics.', 'GOOGLE_ADS_INFO:')
        return ''

    def _build_task_output(self) -> task.TaskOutput:
        if self.task_v1_data['emailAlerts'] and self.task_v1_data['schedule']:
            return task.TaskOutput.EXCLUDE_AND_NOTIFY
        if not self.task_v1_data['emailAlerts'] and self.task_v1_data[
                'schedule']:
            return task.TaskOutput.EXCLUDE
        if self.task_v1_data[
                'emailAlerts'] and not self.task_v1_data['schedule']:
            return task.TaskOutput.NOTIFY
        return task.TaskOutput.EXCLUDE_AND_NOTIFY



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy.orm import mapper
from sqlalchemy.orm import relationship

from googleads_housekeeper.domain.core import execution
from googleads_housekeeper.domain.core import settings
from googleads_housekeeper.domain.core import task
from googleads_housekeeper.domain.external_parsers import website_parser
from googleads_housekeeper.domain.external_parsers import youtube_data_parser
from googleads_housekeeper.domain.placement_handler import entities as placement_entities
from googleads_housekeeper.services import enums

metadata = MetaData()

tasks = Table(
    'tasks',
    metadata,
    Column('_id', Integer, autoincrement=True),
    Column('id', String(36), primary_key=True),
    Column('name', String(255)),
    Column('exclusion_rule', String(255)),
    Column('customer_ids', String(255)),
    Column('date_range', Integer),
    Column('exclusion_level', Enum(enums.ExclusionLevelEnum)),
    Column('output', Enum(task.TaskOutput)),
    Column('from_days_ago', Integer),
    Column('placement_types', String(255)),
    Column('creation_date', DateTime),
    Column('status', Enum(task.TaskStatus)),
    Column('schedule', String(255)),
)

executions = Table('executions', metadata,
                   Column('_id', Integer, autoincrement=True),
                   Column('id', String(36), primary_key=True, nullable=False),
                   Column('task', ForeignKey('tasks.id'), nullable=False),
                   Column('start_time', DateTime, nullable=False),
                   Column('end_time', DateTime, nullable=False),
                   Column('type', Enum(execution.ExecutionTypeEnum)),
                   Column('placements_excluded', Integer))

execution_details = Table(
    'execution_details',
    metadata,
    Column('_id', Integer, autoincrement=True),
    Column('id', String(36), primary_key=True, nullable=False),
    Column('execution_id', ForeignKey('executions.id'), nullable=False),
    Column('placement', String(255), nullable=False),
    Column('placement_type', String(255), nullable=False),
    Column('reason', String(255), nullable=False),
)

allowlisted_placements = Table(
    'allowlisted_placements',
    metadata,
    Column('_id', Integer, autoincrement=True),
    Column('id', String(36), primary_key=True, nullable=False),
    Column('type', Enum(enums.PlacementTypeEnum)),
    Column('name', String(255), unique=True),
    Column('account_id', String(255)),
)

config = Table('config', metadata, Column('_id', Integer, autoincrement=True),
               Column('id', String(36), primary_key=True, nullable=False),
               Column('mcc_id', String(10)),
               Column('always_fetch_youtube_preview_mode', Boolean),
               Column('save_to_db', Boolean),
               Column('email_address', String(50)))

customer_ids = Table('customer_ids', metadata,
                     Column('mcc_id', String(10), primary_key=True),
                     Column('account_name', String(255)),
                     Column('id', String(10), primary_key=True))

mcc_ids = Table('mcc_ids', metadata, Column('id', String(10), primary_key=True),
                Column('account_name', String(255)))

website_info = Table('external_parser.website_info', metadata,
                     Column('_id', Integer, autoincrement=True),
                     Column('id', String(36), primary_key=True, nullable=False),
                     Column('placement', String(255)),
                     Column('title', String(255), nullable=True),
                     Column('description', String(1000), nullable=True),
                     Column('keywords', String(1000), nullable=True),
                     Column('last_processed_time', DateTime),
                     Column('is_processed', Boolean))

youtube_channel_info = Table(
    'external_parser.youtube_channel_info', metadata,
    Column('_id', Integer, autoincrement=True),
    Column('id', String(36), primary_key=True, nullable=False),
    Column('placement', String(255)),
    Column('title', String(255), nullable=True),
    Column('country', String(255), nullable=True),
    Column('description', String(1000), nullable=True),
    Column('viewCount', Integer), Column('subscriberCount', Integer),
    Column('videoCount', Integer),
    Column('topicCategories', String(1000), nullable=True),
    Column('last_processed_time', DateTime), Column('is_processed', Boolean))

youtube_video_info = Table(
    'external_parser.youtube_video_info', metadata,
    Column('_id', Integer, autoincrement=True),
    Column('id', String(36), primary_key=True, nullable=False),
    Column('placement', String(255)),
    Column('title', String(255), nullable=True),
    Column('description', String(1000), nullable=True),
    Column('defaultLanguage', String(10), nullable=True),
    Column('defaultAudioLanguage', String(10), nullable=True),
    Column('commentCount', Integer), Column('favouriteCount', Integer),
    Column('likeCount', Integer), Column('viewCount', Integer),
    Column('madeForKids', Boolean), Column('tags', String(1000)),
    Column('topicCategories', String(1000)),
    Column('last_processed_time', DateTime), Column('is_processed', Boolean))


def start_mappers(engine):
    execution_details_mapper = mapper(execution.ExecutionDetails,
                                      execution_details)
    execution_mapper = mapper(
        execution.Execution,
        executions,
        properties={
            'details':
                relationship(execution_details_mapper, collection_class=list)
        })
    mapper(
        task.Task,
        tasks,
        properties={
            'executions': relationship(execution_mapper, collection_class=list)
        })
    mapper(settings.Config, config)
    mapper(placement_entities.AllowlistedPlacement, allowlisted_placements)
    mapper(settings.MccIds, mcc_ids)
    mapper(settings.CustomerIds, customer_ids)
    mapper(website_parser.WebsiteInfo, website_info)
    mapper(youtube_data_parser.ChannelInfo, youtube_channel_info)
    mapper(youtube_data_parser.VideoInfo, youtube_video_info)
    if engine:
        metadata.create_all(engine)
